#include "haplotypetest.h"

/* Utility function */
extern void          *mem_alloc(size_t, const char *);
extern void           hpsort_reverse_index_of_double_array(int, double *, unsigned int *);

/* Error handler */
extern int            error_check();
extern int            error_message(char *);
extern int            warning_message(char *);

/* Random generator */
extern double         random_number();

/* Frequency initialization */
extern int            Init_frequency(Estimate_Set *, double *, int);

/* Haplotype frequency inference */
extern Estimate_Set  *alloc_estimate_set();
extern int            trait_inheritance_model(Estimate_Set *, int, int);
extern int            no_association_model(Estimate_Set *);

/* EM engine */
extern double         EM_iteration(Estimate_Set *, Option_Data *, double **, double **, double *, int);
extern double         EM_iteration_with_penetrance(Estimate_Set *, Option_Data *, double **, double **, double *, int);
extern double         EM_iteration_with_QTL(Estimate_Set *, Option_Data *, Genotype_Data *, double **, double **, double *, int);

/* Functions for no LD condition */
extern double         noLD_likelihood(Estimate_Set *, double *, int);
extern int            noLD_frequency(Estimate_Set *, double *, double *);
extern int            count_nonzero_haplotype(int, double *, haplotype *);
extern int            estimated_allele_frequency(Estimate_Set *, int, haplotype *, double *, double *);

/* P-value for chi square test */
extern double         calc_p_value_by_chi_square(int, double);

/* Result print */
extern int            print_haplotype_frequency_by_bootstrap(Estimate_Set *, double *, double *, Genotype_Data *);

#ifdef DEBUG_BOOTSTRAP_SAMPLING
extern int            dump_estimate_set(Estimate_Set *);
#endif

int            bootstrap_for_haplotype(Genotype_Data *, Estimate_Set *, Option_Data *);
int            bootstrap_with_penetrance(Genotype_Data *, Estimate_Set *, Option_Data *);
int            bootstrap_with_QTL(Genotype_Data *, Estimate_Set *, Option_Data *);
int            bootstrap_sampling(Genotype_Data *genotype, Genotype_Data *genotype_for_B,
				  Estimate_Set  *table,    Estimate_Set  *table_for_B);

int            bootstrap_for_haplotype(Genotype_Data *genotype,
				       Estimate_Set  *table,
				       Option_Data   *option)
{
  /* Local variables */
  int            b, l, size, init_save, count_max, index_max;
  haplotype      h;
  double         criteria;

  /* Data structure for bootstrap */
  Genotype_Data *genotype_for_B;
  Estimate_Set  *table_for_B;
  double        *f_sum, *s_sum;

  /* Frequencies */
  double        *f_old, *f_new, *f_0, *f_subject, *f_save, *f_tmp;

  /* Parameter for maximum likelihood estimation */
  double        likelihood, likelihood0, LRC, LRC_max;

  /* Result of bootstrap */
  double         rdf, norm;

  /* Message */
  printf("\nBootstrap calculation : number of samples = %d\n",
	 option->bootstrap);

  /* Allocate data structure */
  genotype_for_B           = (Genotype_Data *)mem_alloc(sizeof(Genotype_Data),"Genotype data structure");
  *genotype_for_B          = *genotype;
  genotype_for_B->Subjects = (Subject_Genotype *)mem_alloc(genotype->Number_of_Effective_Subjects*sizeof(Subject_Genotype),"Subject array");

  table_for_B              = alloc_estimate_set();
  *table_for_B             = *table;
  table_for_B->geno_info   = (Genotype_Pack *)mem_alloc(table->number_of_set*sizeof(Genotype_Pack),"Packed genotype structure table");
  table_for_B->target_mask = 0;
  table_for_B->target      = 0;
  table_for_B->order       = (haplotype *)mem_alloc(table->number_of_haplo*sizeof(haplotype),"Array for sort frequency");
  table_for_B->f_allele    = (double *)mem_alloc(table->select_loci*sizeof(double),"Allele frequency for selected loci");

#ifdef DEBUG_BOOTSTRAP_SAMPLING
  for (b=0; b<genotype->Number_of_Subjects; b++) {
    printf("# %5d %s %5d %d\n", b,
	   (genotype->Subjects + b)->Name,
	   (genotype->Subjects + b)->packed_offset,
	   (genotype->Subjects + b)->phenotype);
  }
  dump_estimate_set(table);
#endif

  /* Allocate frequencies and expectation */
  size         = table_for_B->number_of_haplo;
  f_sum        = (double *)mem_alloc(size*sizeof(double),"Frequency summation array");
  s_sum        = (double *)mem_alloc(size*sizeof(double),"Square of frequency summation array");
  f_save       = (double *)mem_alloc(size*sizeof(double),"Frequency array to save");
  f_old        = (double *)mem_alloc(size*sizeof(double),"Frequency array for old step");
  f_new        = (double *)mem_alloc(size*sizeof(double),"Frequency array for new step");
  f_0          = (double *)mem_alloc(size*sizeof(double),"Frequency array for no association");
  f_subject    = (double *)mem_alloc(size*sizeof(double),"Frequency array for each subject");
  for (h=0; h<size; h++) {
    *(f_sum + h) = 0.0;
    *(s_sum + h) = 0.0;
  }

  /* Bootstrap sample loop start */
  for (b=0; b<option->bootstrap; b++) {
    bootstrap_sampling(genotype, genotype_for_B, table, table_for_B);
#ifdef DEBUG_BOOTSTRAP_SAMPLING
    dump_estimate_set(table_for_B);
#endif

    /* EM iteration */
    likelihood  = EM_iteration(table_for_B, option, &f_old, &f_save, f_subject, 0);

    /* Sort haplotype frequency */
    hpsort_reverse_index_of_double_array(table_for_B->number_of_haplo, f_save, table_for_B->order);
    table_for_B->nonzero = count_nonzero_haplotype(table_for_B->number_of_haplo, f_save, table_for_B->order);

    /* Final allele frequency */
    estimated_allele_frequency(table_for_B, table_for_B->nonzero, table_for_B->order, f_save, table_for_B->f_allele);

    /* No LD likelihood */
    noLD_frequency(table_for_B, table_for_B->f_allele, f_0);
    likelihood0 = noLD_likelihood(table_for_B, f_0, 0);
    LRC_max     = likelihood - likelihood0;

    /* Maximum LRC search */
    if (option->maximize > 0) {
      criteria          = log10(option->criteria)*2.0*fabs(LRC_max);
      count_max         = 0;
      index_max         = 0;
      init_save         = option->init_type;
      option->init_type = 2; /* Random distribution */

      /* Loop for LRC search */
      for (l=1; l<=option->maximize; l++){
	/* EM iteration */
	likelihood  = EM_iteration(table_for_B, option, &f_old, &f_new, f_subject, 0);

	/* Sort haplotype frequency */
	hpsort_reverse_index_of_double_array(table_for_B->number_of_haplo, f_new, table_for_B->order);
	table_for_B->nonzero = count_nonzero_haplotype(table_for_B->number_of_haplo, f_new, table_for_B->order);

	/* Final allele frequency */
	estimated_allele_frequency(table_for_B, table_for_B->nonzero, table_for_B->order, f_new, table_for_B->f_allele);

	/* No LD likelihood */
	noLD_frequency(table_for_B, table_for_B->f_allele, f_0);
	likelihood0 = noLD_likelihood(table_for_B, f_0, 0);

	/* Check LRC */
	LRC         = likelihood - likelihood0;
	if      (fabs(LRC - LRC_max) < criteria) {
	  count_max++;
	}
	else if (LRC - LRC_max > criteria) {
	  count_max = 0;
	  index_max = l;
	  LRC_max   = LRC;

	  f_tmp     = f_save;
	  f_save    = f_new;
	  f_new     = f_tmp;
#ifdef DEBUG_BOOTSTRAP
	  printf("Restart No. %5d : LRC maximum                     = %13.6e\n",l,LRC);
#endif
	}
      }

#ifdef DEBUG_BOOTSTRAP
      printf("  Final LRC                                         = %13.6e at %d\n",LRC_max,index_max);
      printf("  Same LRC came out %d times\n",count_max);
#endif
    
      option->init_type = init_save;
    }

    if (option->message) {
      printf("Bootstrap No. %5d\n",b);
      printf("  log10 likelihood assuming no linkage disequilibrium = %13.6e\n",likelihood);
      printf("  Final log10 Likelihood                              = %13.6e\n",likelihood);
      printf("  Likelihood Ratio Criteria (LRC)                     = %13.6e\n",LRC_max);
    }

    /* Save result */
    for (h=0; h<size; h++) {
      *(f_sum + h) += *(f_save + h);
      *(s_sum + h) += *(f_save + h) * *(f_save + h);
    }
  }

#ifdef DEBUG_BOOTSTRAP
  for (h=0; h<size; h++) {
    printf("# haplotype = %5d, f = %f, f^2 = %f, f*f/n = %f\n",
	   h,*(f_sum + h),*(s_sum + h),*(f_sum + h) * *(f_sum + h)/(double)option->bootstrap);
  }
#endif
  norm = 1.0 / (double)option->bootstrap;
  rdf  = 1.0 / (double)(option->bootstrap - 1);
  for (h=0; h<size; h++) {
    *(s_sum + h) = *(s_sum + h) - *(f_sum + h) * *(f_sum + h) * norm;
    if (*(s_sum + h) < 0.0) {
      *(s_sum + h) = 0.0;
    }
    *(s_sum + h) = sqrt(*(s_sum + h) * rdf);
    *(f_sum + h) = *(f_sum + h) * norm;
  }
#ifdef DEBUG_BOOTSTRAP
  for (h=0; h<size; h++) {
    printf("# haplotype = %5d, average of f = %f, SE of f = %f\n",
	   h,*(f_sum + h),*(s_sum + h));
  }
#endif

  table_for_B->f = f_sum;
  hpsort_reverse_index_of_double_array(table_for_B->number_of_haplo, table_for_B->f, table_for_B->order);
  table->nonzero = count_nonzero_haplotype(table_for_B->number_of_haplo, table_for_B->f, table_for_B->order);
  print_haplotype_frequency_by_bootstrap(table_for_B, table->f, s_sum, genotype);

  free(f_subject);
  free(f_0);
  free(f_new);
  free(f_old);
  free(f_save);
  free(s_sum);
  free(f_sum);
  free(table_for_B->f_allele);
  free(table_for_B->order);
  free(table_for_B->geno_info);
  free(table_for_B);
  free(genotype_for_B->Subjects);
  free(genotype_for_B);

  return 0;
}

int            bootstrap_with_penetrance(Genotype_Data *genotype,
					 Estimate_Set  *table,
					 Option_Data   *option)
{
  /* Local variables */
  int            b, l, size, init_save;
  int            count_max, index_max, count0_max, index0_max;
  haplotype      h;
  double         conv = 2.0 * log(10.0);
  double         s, s_upper, s_lower, p_value, p_upper, p_lower;
  double         criteria;

  /* Data structure for bootstrap */
  Genotype_Data *genotype_for_B;
  Estimate_Set  *table_for_B;
  double        *f_sum, *s_sum;

  /* Frequencies */
  double        *f_old, *f_new, *f_0, *f_subject, *f_save, *f_tmp;

  /* Parameter for maximum likelihood estimation */
  double         likelihood, likelihood0, LRC_max;
  double         likelihood_max, likelihood0_max;

  /* Result of bootstrap */
  double         p_plus[3], p_plus_s[3];
  double         pt_sum[3], st_sum[3];
  double         rdf, norm;
  double         ps_sum, ss_sum, typeII;
  int            p_count;

  /* Message */
  printf("\nBootstrap calculation : number of samples = %d\n",
	 option->bootstrap);

  /* Allocate data structure */
  genotype_for_B           = (Genotype_Data *)mem_alloc(sizeof(Genotype_Data),"Genotype data structure");
  *genotype_for_B          = *genotype;
  genotype_for_B->Subjects = (Subject_Genotype *)mem_alloc(genotype->Number_of_Effective_Subjects*sizeof(Subject_Genotype),"Subject array");

  table_for_B              = alloc_estimate_set();
  *table_for_B             = *table;
  table_for_B->geno_info   = (Genotype_Pack *)mem_alloc(table->number_of_set*sizeof(Genotype_Pack),"Packed genotype structure table");
  table_for_B->order       = (haplotype *)mem_alloc(table->number_of_haplo*sizeof(haplotype),"Array for sort frequency");

#ifdef DEBUG_BOOTSTRAP_SAMPLING
  for (b=0; b<genotype->Number_of_Subjects; b++) {
    printf("# %5d %s %5d %d\n", b,
	   (genotype->Subjects + b)->Name,
	   (genotype->Subjects + b)->packed_offset,
	   (genotype->Subjects + b)->phenotype);
  }
  dump_estimate_set(table);
#endif

  /* Allocate frequencies and expectation */
  size         = table_for_B->number_of_haplo;
  f_sum        = (double *)mem_alloc(size*sizeof(double),"Frequency summation array");
  s_sum        = (double *)mem_alloc(size*sizeof(double),"Square of frequency summation array");
  f_save       = (double *)mem_alloc(size*sizeof(double),"Frequency array to save");
  f_old        = (double *)mem_alloc(size*sizeof(double),"Frequency array for old step");
  f_new        = (double *)mem_alloc(size*sizeof(double),"Frequency array for new step");
  f_0          = (double *)mem_alloc(size*sizeof(double),"Frequency array for no association");
  f_subject    = (double *)mem_alloc(size*sizeof(double),"Frequency array for each subject");
  for (h=0; h<size; h++) {
    *(f_sum + h) = 0.0;
    *(s_sum + h) = 0.0;
  }
  for (b=0; b<3; b++) {
    pt_sum[b] = 0.0;
    st_sum[b] = 0.0;
  }
  ps_sum = 0.0;
  ss_sum = 0.0;
  p_count = 0;

  /* Bootstrap sample loop start */
  for (b=0; b<option->bootstrap; b++) {
    bootstrap_sampling(genotype, genotype_for_B, table, table_for_B);
#ifdef DEBUG_BOOTSTRAP_SAMPLING
    dump_estimate_set(table_for_B);
#endif

    /* Initialize probability */
    Init_frequency(table, f_save, option->init_type);
    for (h=0; h<table->number_of_haplo; h++) {
      *(f_0 + h) = *(f_save + h);
    }

    /* EM iteration */
    trait_inheritance_model(table_for_B, table_for_B->model_type, 0);
    likelihood  = EM_iteration_with_penetrance(table_for_B, option, &f_old, &f_save, f_subject, 0);
    p_plus[0]   = table_for_B->p_plus[0];
    p_plus[1]   = table_for_B->p_plus[1];
    p_plus[2]   = table_for_B->p_plus[2];

    /* EM iteration for no association */
    no_association_model(table_for_B);
    likelihood0 = EM_iteration_with_penetrance(table_for_B, option, &f_old, &f_0,    f_subject, 0);
    LRC_max         = likelihood - likelihood0;
    likelihood_max  = likelihood;
    likelihood0_max = likelihood0;

    /* Maximum LRC search */
    if (option->maximize > 0) {
      criteria          = log10(option->criteria)*2.0*fabs(LRC_max);
      count_max         = 0;
      index_max         = 0;
      count0_max        = 0;
      index0_max        = 0;
      init_save         = option->init_type;
      option->init_type = 2; /* Random distribution */

      /* Loop for LRC search */
      for (l=1; l<=option->maximize; l++){
	/* Initialize probability */
	Init_frequency(table, f_new, option->init_type);
	for (h=0; h<table->number_of_haplo; h++) {
	  *(f_0 + h) = *(f_new + h);
	}

	/* EM iteration */
	trait_inheritance_model(table_for_B, table_for_B->model_type, 0);
	likelihood  = EM_iteration_with_penetrance(table_for_B, option, &f_old, &f_new, f_subject, 0);
	p_plus_s[0] = table_for_B->p_plus[0];
	p_plus_s[1] = table_for_B->p_plus[1];
	p_plus_s[2] = table_for_B->p_plus[2];

	/* EM iteration for no association */
	no_association_model(table_for_B);
	likelihood0 = EM_iteration_with_penetrance(table_for_B, option, &f_old, &f_0, f_subject, 0);

	/* Check Likelihood */
	if      (fabs(likelihood - likelihood_max) < criteria) {
	  count_max++;
	}
	else if (likelihood - likelihood_max > criteria) {
	  count_max       = 0;
	  index_max       = l;
	  likelihood_max  = likelihood;

	  f_tmp     = f_save;
	  f_save    = f_new;
	  f_new     = f_tmp;

	  p_plus[0] = p_plus_s[0];
	  p_plus[1] = p_plus_s[1];
	  p_plus[2] = p_plus_s[2];
#ifdef DEBUG_BOOTSTRAP
	  printf("Restart No. %5d : Likelihood for alternative      = %13.6e\n",l,likelihood);
#endif
	}

	if      (fabs(likelihood0 - likelihood0_max) < criteria) {
	  count0_max++;
	}
	else if (likelihood0 - likelihood0_max > criteria) {
	  count0_max      = 0;
	  index0_max      = l;
	  likelihood0_max = likelihood0;

#ifdef DEBUG_BOOTSTRAP
	  printf("Restart No. %5d : Likelihood for null             = %13.6e\n",l,likelihood0);
#endif
	}
      }
      LRC_max = likelihood_max - likelihood0_max;
#ifdef DEBUG_BOOTSTRAP
      printf("  Final Likelihood for alternative hypothesis       = %13.6e at %d\n",likelihood_max,index_max);
      printf("  Same Likelihood came out %d times\n",count_max);
      printf("  Final Likelihood for null hypothesis              = %13.6e at %d\n",likelihood0_max,index0_max);
      printf("  Same Likelihood came out %d times\n",count0_max);
      printf("  Final LRC                                         = %13.6e\n",LRC_max);
#endif
    
      option->init_type = init_save;
    }

    s           = conv * LRC_max;
    if (s <= 0.0) {
      s = 0.0;
    }
    p_value     = calc_p_value_by_chi_square(1, s);

    if (p_value >= option->s_level) {
      p_count++;
    }

    if (option->message) {
      printf("Bootstrap No. %5d\n",b);
      printf("  log10 likelihood assuming no association = %13.6e\n",likelihood0);
      printf("  Final log10 Likelihood                   = %13.6e\n",likelihood);
      printf("  Likelihood Ratio Criteria (LRC)          = %13.6e\n",LRC_max);
      printf("  Penetrance for no target haplotype       = %7.4f\n",p_plus[0]);
      printf("  Penetrance for one target haplotype      = %7.4f\n",p_plus[1]);
      printf("  Penetrance for two target haplotype      = %7.4f\n",p_plus[2]);
      printf("  Statistic for chi-square test by LRC     = %.4f\n",s);
      printf("                   degree_of_freedom       = 1\n");
      printf("                             p_value       = %.4e\n",p_value);
    }

    /* Save result */
    for (h=0; h<size; h++) {
      *(f_sum + h) += *(f_save + h);
      *(s_sum + h) += *(f_save + h) * *(f_save + h);
    }
    pt_sum[0] += p_plus[0];
    pt_sum[1] += p_plus[1];
    pt_sum[2] += p_plus[2];
    st_sum[0] += p_plus[0]*p_plus[0];
    st_sum[1] += p_plus[1]*p_plus[1];
    st_sum[2] += p_plus[2]*p_plus[2];
    ps_sum += s;
    ss_sum += s * s;
  }

#ifdef DEBUG_BOOTSTRAP
  for (h=0; h<size; h++) {
    printf("# haplotype = %5d, f = %f, f^2 = %f, f*f/n = %f\n",
	   h,*(f_sum + h),*(s_sum + h),*(f_sum + h) * *(f_sum + h)/(double)option->bootstrap);
  }
  printf("# penetrance = %f %f %f %f %f %f\n",pt_sum[0],st_sum[0],pt_sum[1],st_sum[1],pt_sum[2],st_sum[2]);
  printf("# s          = %f %f %d\n",ps_sum,ss_sum,p_count);
#endif
  norm = 1.0 / (double)option->bootstrap;
  rdf  = 1.0 / (double)(option->bootstrap - 1);
  for (h=0; h<size; h++) {
    *(s_sum + h) = *(s_sum + h) - *(f_sum + h) * *(f_sum + h) * norm;
    if (*(s_sum + h) < 0.0) {
      *(s_sum + h) = 0.0;
    }
    *(s_sum + h) = sqrt(*(s_sum + h) * rdf);
    *(f_sum + h) = *(f_sum + h) * norm;
  }

  for (b=0; b<3; b++) {
    st_sum[b] = st_sum[b] - pt_sum[b] * pt_sum[b] * norm;
    if (st_sum[b] < 0.0) {
      st_sum[b] = 0.0;
    }
    st_sum[b] = sqrt(st_sum[b] * rdf);
    pt_sum[b] = pt_sum[b] * norm;
  }

  ss_sum = ss_sum - ps_sum * ps_sum * norm;
  if (ss_sum < 0.0) {
    ss_sum = 0.0;
  }
  ss_sum = sqrt(ss_sum * rdf);
  ps_sum = ps_sum * norm;

  s       = ps_sum;
  s_lower = ps_sum - ss_sum;
  s_upper = ps_sum + ss_sum;
  if (s_lower <= 0.0) {
    s_lower = 0.0;
  }
  p_value = calc_p_value_by_chi_square(table->df_for_test, s);
  p_lower = calc_p_value_by_chi_square(table->df_for_test, s_lower);
  p_upper = calc_p_value_by_chi_square(table->df_for_test, s_upper);
  typeII  = (double)p_count * norm;
#ifdef DEBUG_BOOTSTRAP
  for (h=0; h<size; h++) {
    printf("# haplotype = %5d, average of f = %f, SE of f = %f\n",
	   h,*(f_sum + h),*(s_sum + h));
  }
  printf("# penetrance = %f %f %f %f %f %f\n",pt_sum[0],st_sum[0],pt_sum[1],st_sum[1],pt_sum[2],st_sum[2]);
  printf("# s          = %f %f %f %f %f %d\n",ps_sum,ss_sum,p_value,p_lower,p_upper,p_count);
#endif

  table_for_B->f = f_sum;
  hpsort_reverse_index_of_double_array(table_for_B->number_of_haplo, table_for_B->f, table_for_B->order);
  table->nonzero = count_nonzero_haplotype(table_for_B->number_of_haplo, table_for_B->f, table_for_B->order);
  print_haplotype_frequency_by_bootstrap(table_for_B, table->f, s_sum, genotype);

  printf("\nEstimated penetrance by non-parametric bootstrap\n");
  printf("                        estimated    average         SE\n");
  if (table->model_type == GENOTYPE_MODEL) {
    printf("  no target haplotype  %10.4f %10.4f %10.4f\n",table->p_plus[0],pt_sum[0],st_sum[0]);
    printf("  1 target haplotype   %10.4f %10.4f %10.4f\n",table->p_plus[1],pt_sum[1],st_sum[1]);
    printf("  2 target haplotypes  %10.4f %10.4f %10.4f\n",table->p_plus[2],pt_sum[2],st_sum[2]);
  }
  else {
    printf("  target haplotype     %10.4f %10.4f %10.4f\n",table->p_plus[1],pt_sum[1],st_sum[1]);
    printf("  the other haplotypes %10.4f %10.4f %10.4f\n",table->p_plus[0],pt_sum[0],st_sum[0]);
  }
  printf("\nStatistic for chi-square test by bootstrap\n");
  printf("                          average         SE average-SE average+SE\n");
  printf("  statistics           %10.4f %10.4f %10.4f %10.4f\n",ps_sum,ss_sum,s_lower,s_upper);
  printf("  p_value              %10.4e            %10.4e %10.4e\n",p_value,p_lower,p_upper);
  printf("\n  typeII error by bootstrap sampling = %7.4f (significance level = %7.4f)\n",typeII,option->s_level);

  trait_inheritance_model(table, table->model_type, 0);
  free(f_subject);
  free(f_0);
  free(f_new);
  free(f_old);
  free(f_save);
  free(s_sum);
  free(f_sum);
  free(table_for_B->order);
  free(table_for_B->geno_info);
  free(table_for_B);
  free(genotype_for_B->Subjects);
  free(genotype_for_B);

  return 0;
}

int            bootstrap_with_QTL(Genotype_Data *genotype,
				  Estimate_Set  *table,
				  Option_Data   *option)
{
  /* Local variables */
  int            b, l, size, init_save;
  int            count_max, index_max, count0_max, index0_max;
  haplotype      h;
  double         conv = 2.0 * log(10.0);
  double         s, s_upper, s_lower, p_value, p_upper, p_lower;
  double         criteria;

  /* Data structure for bootstrap */
  Genotype_Data *genotype_for_B;
  Estimate_Set  *table_for_B;
  double        *f_sum, *s_sum;

  /* Frequencies */
  double        *f_old, *f_new, *f_0, *f_subject, *f_save, *f_tmp;

  /* Parameter for maximum likelihood estimation */
  double         likelihood, likelihood0, LRC_max;
  double         likelihood_max, likelihood0_max;

  /* Result of bootstrap */
  double         mu[3],   sigma,   sigma_c[3];
  double         mu_s[3], sigma_s, sigma_c_s[3];
  double         statistic[8][2];
  /* statistic[0][*] : mu[0]      & mu_[0]^2
     statistic[1][*] : mu[1]      & mu_[1]^2
     statistic[2][*] : mu[2]      & mu_[2]^2
     statistic[3][*] : sigma      & sigma^2
     statistic[4][*] : sigma_c[0] & sigma_c[0]^2
     statistic[5][*] : sigma_c[1] & sigma_c[1]^2
     statistic[6][*] : sigma_c[2] & sigma_c[2]^2
     statistic[7][*] : s          & s^2
   */
  double         rdf, norm, typeII;
  int            p_count;

  /* Message */
  printf("\nBootstrap calculation : number of samples = %d\n",
	 option->bootstrap);

  /* Allocate data structure */
  genotype_for_B           = (Genotype_Data *)mem_alloc(sizeof(Genotype_Data),"Genotype data structure");
  *genotype_for_B          = *genotype;
  genotype_for_B->Subjects = (Subject_Genotype *)mem_alloc(genotype->Number_of_Effective_Subjects*sizeof(Subject_Genotype),"Subject array");

  table_for_B              = alloc_estimate_set();
  *table_for_B             = *table;
  table_for_B->geno_info   = (Genotype_Pack *)mem_alloc(table->number_of_set*sizeof(Genotype_Pack),"Packed genotype structure table");
  table_for_B->order       = (haplotype *)mem_alloc(table->number_of_haplo*sizeof(haplotype),"Array for sort frequency");

#ifdef DEBUG_BOOTSTRAP_SAMPLING
  for (b=0; b<genotype->Number_of_Subjects; b++) {
    printf("# %5d %s %5d %d %f\n", b,
	   (genotype->Subjects + b)->Name,
	   (genotype->Subjects + b)->packed_offset,
	   (genotype->Subjects + b)->phenotype,
	   (genotype->Subjects + b)->QTL);
  }
  dump_estimate_set(table);
#endif

  /* Allocate frequencies and expectation */
  size         = table_for_B->number_of_haplo;
  f_sum        = (double *)mem_alloc(size*sizeof(double),"Frequency summation array");
  s_sum        = (double *)mem_alloc(size*sizeof(double),"Square of frequency summation array");
  f_save       = (double *)mem_alloc(size*sizeof(double),"Frequency array to save");
  f_old        = (double *)mem_alloc(size*sizeof(double),"Frequency array for old step");
  f_new        = (double *)mem_alloc(size*sizeof(double),"Frequency array for new step");
  f_0          = (double *)mem_alloc(size*sizeof(double),"Frequency array for no association");
  f_subject    = (double *)mem_alloc(size*sizeof(double),"Frequency array for each subject");
  for (h=0; h<size; h++) {
    *(f_sum + h) = 0.0;
    *(s_sum + h) = 0.0;
  }
  for (b=0; b<6; b++) {
    statistic[b][0] = 0.0;
    statistic[b][1] = 0.0;
  }
  p_count = 0;

  /* Bootstrap sample loop start */
  for (b=0; b<option->bootstrap; b++) {
    bootstrap_sampling(genotype, genotype_for_B, table, table_for_B);
#ifdef DEBUG_BOOTSTRAP_SAMPLING
    dump_estimate_set(table_for_B);
#endif

    /* Initialize probability */
    Init_frequency(table, f_save, option->init_type);
    for (h=0; h<table->number_of_haplo; h++) {
      *(f_0 + h) = *(f_save + h);
    }

    /* EM iteration */
    trait_inheritance_model(table_for_B, table_for_B->model_type, 0);
    likelihood  = EM_iteration_with_QTL(table_for_B, option, genotype_for_B, &f_old, &f_save, f_subject, 0);
    mu[0]      = table_for_B->mu[0];
    mu[1]      = table_for_B->mu[1];
    mu[2]      = table_for_B->mu[2];
    sigma      = table_for_B->SD;
    sigma_c[0] = table_for_B->SD_c[0];
    sigma_c[1] = table_for_B->SD_c[1];
    sigma_c[2] = table_for_B->SD_c[2];

    /* EM iteration for no association */
    no_association_model(table_for_B);
    likelihood0 = EM_iteration_with_QTL(table_for_B, option, genotype_for_B, &f_old, &f_0,    f_subject, 0);
    LRC_max     = likelihood - likelihood0;
    likelihood_max  = likelihood;
    likelihood0_max = likelihood0;

    /* Maximum LRC search */
    if (option->maximize > 0) {
      criteria          = log10(option->criteria)*2.0*fabs(LRC_max);
      count_max         = 0;
      index_max         = 0;
      count0_max        = 0;
      index0_max        = 0;
      init_save         = option->init_type;
      option->init_type = 2; /* Random distribution */

      /* Loop for LRC search */
      for (l=1; l<=option->maximize; l++){
	/* Initialize probability */
	Init_frequency(table, f_new, option->init_type);
	for (h=0; h<table->number_of_haplo; h++) {
	  *(f_0 + h) = *(f_new + h);
	}

	/* EM iteration */
	trait_inheritance_model(table_for_B, table_for_B->model_type, 0);
	likelihood    = EM_iteration_with_QTL(table_for_B, option, genotype_for_B, &f_old, &f_new, f_subject, 0);
	mu_s[0]       = table_for_B->mu[0];
	mu_s[1]       = table_for_B->mu[1];
	mu_s[2]       = table_for_B->mu[2];
	sigma_s       = table_for_B->SD;
	sigma_c_s[0]  = table_for_B->SD_c[0];
	sigma_c_s[1]  = table_for_B->SD_c[1];
	sigma_c_s[2]  = table_for_B->SD_c[2];

	/* EM iteration for no association */
	no_association_model(table_for_B);
	likelihood0   = EM_iteration_with_QTL(table_for_B, option, genotype_for_B, &f_old, &f_0,    f_subject, 0);

	/* Check Likelihood */
	if      (fabs(likelihood - likelihood_max) < criteria) {
	  count_max++;
	}
	else if (likelihood - likelihood_max > criteria) {
	  count_max       = 0;
	  index_max       = l;
	  likelihood_max  = likelihood;

	  f_tmp       = f_save;
	  f_save      = f_new;
	  f_new       = f_tmp;

	  mu[0]      = mu_s[0];
	  mu[1]      = mu_s[1];
	  mu[2]      = mu_s[2];
	  sigma      = sigma_s;
	  sigma_c[0] = sigma_c_s[0];
	  sigma_c[1] = sigma_c_s[1];
	  sigma_c[2] = sigma_c_s[2];
#ifdef DEBUG_BOOTSTRAP
	  printf("Restart No. %5d : Likelihood for alternative      = %13.6e\n",l,likelihood);
#endif
	}

	if      (fabs(likelihood0 - likelihood0_max) < criteria) {
	  count0_max++;
	}
	else if (likelihood0 - likelihood0_max > criteria) {
	  count0_max      = 0;
	  index0_max      = l;
	  likelihood0_max = likelihood0;

#ifdef DEBUG_BOOTSTRAP
	  printf("Restart No. %5d : Likelihood for null             = %13.6e\n",l,likelihood0);
#endif
	}
      }
      LRC_max = likelihood_max - likelihood0_max;
#ifdef DEBUG_BOOTSTRAP
      printf("  Final Likelihood for alternative hypothesis       = %13.6e at %d\n",likelihood_max,index_max);
      printf("  Same Likelihood came out %d times\n",count_max);
      printf("  Final Likelihood for null hypothesis              = %13.6e at %d\n",likelihood0_max,index0_max);
      printf("  Same Likelihood came out %d times\n",count0_max);
      printf("  Final LRC                                         = %13.6e\n",LRC_max);
#endif
    
      option->init_type = init_save;
    }

    s           = conv * LRC_max;
    if (s <= 0.0) {
      s = 0.0;
    }
    p_value     = calc_p_value_by_chi_square(1, s);

    if (p_value >= option->s_level) {
      p_count++;
    }

    if (option->message) {
      printf("Bootstrap No. %5d\n",b);
      printf("  log10 likelihood assuming no association = %13.6e\n",likelihood0);
      printf("  Final log10 Likelihood                   = %13.6e\n",likelihood);
      printf("  Likelihood Ratio Criteria (LRC)          = %13.6e\n",LRC_max);
      printf("  Mu for no target haplotype               = %7.4f\n",mu[0]);
      printf("  Mu for 1 target haplotype                = %7.4f\n",mu[1]);
      printf("  Mu for 2 target haplotypes               = %7.4f\n",mu[2]);
      printf("  Statistic for chi-square test by LRC     = %.4f\n",s);
      printf("                   degree_of_freedom       = 1\n");
      printf("                             p_value       = %.4e\n",p_value);
    }

    /* Save result */
    for (h=0; h<size; h++) {
      *(f_sum + h) += *(f_save + h);
      *(s_sum + h) += *(f_save + h) * *(f_save + h);
    }
    statistic[0][0] += mu[0];
    statistic[0][1] += mu[0] * mu[0];
    statistic[1][0] += mu[1];
    statistic[1][1] += mu[1] * mu[1];
    statistic[2][0] += mu[2];
    statistic[2][1] += mu[2] * mu[2];
    statistic[3][0] += sigma;
    statistic[3][1] += sigma * sigma;
    statistic[4][0] += sigma_c[0];
    statistic[4][1] += sigma_c[0] * sigma_c[0];
    statistic[5][0] += sigma_c[1];
    statistic[5][1] += sigma_c[1] * sigma_c[1];
    statistic[6][0] += sigma_c[2];
    statistic[6][1] += sigma_c[2] * sigma_c[2];
    statistic[7][0] += s;
    statistic[7][1] += s * s;
  }

#ifdef DEBUG_BOOTSTRAP
  for (h=0; h<size; h++) {
    printf("# haplotype = %5d, f = %f, f^2 = %f, f*f/n = %f\n",
	   h,*(f_sum + h),*(s_sum + h),*(f_sum + h) * *(f_sum + h)/(double)option->bootstrap);
  }
  printf("# mu         = %f %f %f %f %f %f\n",
	 statistic[0][0],statistic[0][1],statistic[1][0],statistic[1][1],statistic[2][0],statistic[2][1]);
  printf("# sigma      = %f %f %f %f %f %f %f %f\n",
	 statistic[3][0],statistic[3][1],statistic[4][0],statistic[4][1],
	 statistic[5][0],statistic[5][1],statistic[6][0],statistic[6][1]);
  printf("# s          = %f %f %d\n",statistic[7][0],statistic[7][1],p_count);
#endif
  norm = 1.0 / (double)option->bootstrap;
  rdf  = 1.0 / (double)(option->bootstrap - 1);
  for (h=0; h<size; h++) {
    *(s_sum + h) = *(s_sum + h) - *(f_sum + h) * *(f_sum + h) * norm;
    if (*(s_sum + h) < 0.0) {
      *(s_sum + h) = 0.0;
    }
    *(s_sum + h) = sqrt(*(s_sum + h) * rdf);
    *(f_sum + h) = *(f_sum + h) * norm;
  }

  for (b=0; b<8; b++) {
    statistic[b][1] = statistic[b][1] - statistic[b][0] * statistic[b][0] * norm;
    if (statistic[b][1] < 0.0) {
      statistic[b][1] = 0.0;
    }
    statistic[b][1] = sqrt(statistic[b][1] * rdf);
    statistic[b][0] = statistic[b][0] * norm;
  }

  s       = statistic[7][0];
  s_lower = statistic[7][0] - statistic[7][1];
  s_upper = statistic[7][0] + statistic[7][1];
  if (s_lower <= 0.0) {
    s_lower = 0.0;
  }
  p_value = calc_p_value_by_chi_square(1, s);
  p_lower = calc_p_value_by_chi_square(1, s_lower);
  p_upper = calc_p_value_by_chi_square(1, s_upper);
  typeII  = (double)p_count * norm;
#ifdef DEBUG_BOOTSTRAP
  for (h=0; h<size; h++) {
    printf("# haplotype = %5d, average of f = %f, SE of f = %f\n",
	   h,*(f_sum + h),*(s_sum + h));
  }
  printf("# mu         = %f %f %f %f\n",
	 statistic[0][0],statistic[0][1],statistic[1][0],statistic[1][1],statistic[2][0],statistic[2][1]);
  printf("# sigma      = %f %f %f %f %f %f %f %f\n",
	 statistic[3][0],statistic[3][1],statistic[4][0],statistic[4][1],
	 statistic[5][0],statistic[5][1],statistic[6][0],statistic[6][1]);
  printf("# s          = %f %f %f %f %f %d\n",statistic[7][0],statistic[7][1],p_value,p_lower,p_upper,p_count);
#endif

  table_for_B->f = f_sum;
  hpsort_reverse_index_of_double_array(table_for_B->number_of_haplo, table_for_B->f, table_for_B->order);
  table->nonzero = count_nonzero_haplotype(table_for_B->number_of_haplo, table_for_B->f, table_for_B->order);
  print_haplotype_frequency_by_bootstrap(table_for_B, table->f, s_sum, genotype);

  printf("\nEstimated QTL parameters by non-parametric bootstrap\n");
  printf("                        estimated    average         SE\n");
  if (table->model_type == GENOTYPE_MODEL) {
    printf("  mean with no target  %10.4f %10.4f %10.4f\n",table->mu[0],  statistic[0][0],statistic[0][1]);
    printf("  mean with 1 target   %10.4f %10.4f %10.4f\n",table->mu[1],  statistic[1][0],statistic[1][1]);
    printf("  mean with 2 targets  %10.4f %10.4f %10.4f\n",table->mu[2],  statistic[2][0],statistic[2][1]);
    printf("  sigma                %10.4f %10.4f %10.4f\n",table->SD,     statistic[3][0],statistic[3][1]);
    printf("  sigma with no target %10.4f %10.4f %10.4f\n",table->SD_c[0],statistic[4][0],statistic[4][1]);
    printf("  sigma with 1 target  %10.4f %10.4f %10.4f\n",table->SD_c[1],statistic[5][0],statistic[5][1]);
    printf("  sigma with 2 targets %10.4f %10.4f %10.4f\n",table->SD_c[2],statistic[6][0],statistic[6][1]);
  }
  else {
    printf("  mean+                %10.4f %10.4f %10.4f\n",table->mu[1],  statistic[1][0],statistic[1][1]);
    printf("  mean-                %10.4f %10.4f %10.4f\n",table->mu[0],  statistic[0][0],statistic[0][1]);
    printf("  sigma                %10.4f %10.4f %10.4f\n",table->SD,     statistic[3][0],statistic[3][1]);
    printf("  sigma+               %10.4f %10.4f %10.4f\n",table->SD_c[0],statistic[4][0],statistic[4][1]);
    printf("  sigma-               %10.4f %10.4f %10.4f\n",table->SD_c[1],statistic[5][0],statistic[5][1]);
  }
  printf("\nStatistic for chi-square test by bootstrap\n");
  printf("                          average         SE average-SE average+SE\n");
  printf("  statistics           %10.4f %10.4f %10.4f %10.4f\n",statistic[7][0],statistic[7][1],s_lower,s_upper);
  printf("  p_value              %10.4e            %10.4e %10.4e\n",p_value,p_lower,p_upper);
  printf("\n  typeII error by bootstrap sampling = %7.4f (significance level = %7.4f)\n",typeII,option->s_level);


  trait_inheritance_model(table, 1, 0);
  free(f_subject);
  free(f_0);
  free(f_new);
  free(f_old);
  free(f_save);
  free(s_sum);
  free(f_sum);
  free(table_for_B->order);
  free(table_for_B->geno_info);
  free(table_for_B);
  free(genotype_for_B->Subjects);
  free(genotype_for_B);

  return 0;
}

int            bootstrap_sampling(Genotype_Data *genotype,
				  Genotype_Data *genotype_for_B,
				  Estimate_Set  *table,
				  Estimate_Set  *table_for_B)
{
  /* Local variables */
  double       r;
  int          i, s, j;
  int          number_of_set, offset;
  int          not_match;

  /* Reset counter and pointers */
  number_of_set = 0;
  i             = 0;

  /* Bootstrap sampling */
  while (i<genotype_for_B->Number_of_Effective_Subjects) {
    r = random_number();
    s = (int)(r * (double)genotype_for_B->Number_of_Subjects);
    if (s >= genotype_for_B->Number_of_Subjects) {
      s--;
    }

    offset = (genotype->Subjects + s)->packed_offset;
#ifdef DEBUG_BOOTSTRAP_SAMPLING
    printf("# %5d %5d %f %5d %5d : ",i,genotype_for_B->Number_of_Effective_Subjects,r,s,offset);
#endif

    if (offset >= 0) {
      *(genotype_for_B->Subjects + i) = *(genotype->Subjects + s);
      if (number_of_set == 0) {
	*(table_for_B->geno_info + number_of_set) = *(table->geno_info + offset);
	(table_for_B->geno_info  + number_of_set)->weight = 1.0;
	(genotype_for_B->Subjects + i)->packed_offset = number_of_set;
	number_of_set++;
      }
      else {
	not_match = 1;
	for (j=0; j<number_of_set; j++) {
	  if (((table->geno_info + offset)->Genotype_Code[0] == (table_for_B->geno_info + j)->Genotype_Code[0]) &&
	      ((table->geno_info + offset)->Genotype_Code[1] == (table_for_B->geno_info + j)->Genotype_Code[1]) &&
	      ((table->geno_info + offset)->Missing[0]       == (table_for_B->geno_info + j)->Missing[0]      ) &&
	      ((table->geno_info + offset)->Missing[1]       == (table_for_B->geno_info + j)->Missing[1]      ) &&
	      ((table->geno_info + offset)->phenotype        == (table_for_B->geno_info + j)->phenotype       )) {
	    not_match = 0;
	    (table_for_B->geno_info   + j)->weight += 1.0;
	    (genotype_for_B->Subjects + i)->packed_offset = j;
	  }
	}
	if (not_match) {
	  *(table_for_B->geno_info + number_of_set) = *(table->geno_info + offset);
	  (table_for_B->geno_info  + number_of_set)->weight = 1.0;
	  (genotype_for_B->Subjects + i)->packed_offset = number_of_set;
	  number_of_set++;
	}
      }
#ifdef DEBUG_BOOTSTRAP_SAMPLING
      printf("  %5d %s %5d %d %f\n",i,(genotype_for_B->Subjects + i)->Name,
	     (genotype_for_B->Subjects + i)->packed_offset,
	     (genotype_for_B->Subjects + i)->phenotype,
	     (genotype_for_B->Subjects + i)->QTL);
#endif
      i++;
    }
#ifdef DEBUG_BOOTSTRAP_SAMPLING
    else {
      printf("\n");
    }
#endif
  }

  table_for_B->number_of_set = number_of_set;

  return 0;
}
