#include "haplotypetest.h"

/* Utility function */
extern void          *mem_alloc(size_t, const char *);

/* Error handler */
extern int            error_check();
extern int            error_message(char *);
extern int            warning_message(char *);

/* Random generator */
extern double         random_number();

#ifdef DEBUG
extern char          *binary_string(haplotype);
#endif

/* Define functions */
double         EM_iteration(Estimate_Set *, Option_Data *, double **, double **, double *, int);
int            Init_frequency(Estimate_Set *, double *, int);
double         Total_likelihood(Estimate_Set *, double *);
double         Subject_likelihood(Estimate_Set *, double *, double *);
double         Pop_frequency(Estimate_Set *, double *, double *, double *);
double         diplotype_probability(int, Diplotype *, double *, double *);

double         EM_iteration(Estimate_Set  *table,
			    Option_Data   *option,
			    double       **f1, /* f_old */
			    double       **f2, /* f_new */
			    double        *f_subject,
			    int            message)
{
  /* Frequencies */
  double     *f_old, *f_new, *f_tmp;

  /* EM calculation parameter */
  double      likelihood, likeli_old;
  double      criteria, change;
  double      delta_f, delta_f_max, delta_f_sum;
  int         loop;
  haplotype   h;

  f_old    = *f1;
  f_new    = *f2;
  criteria = log10(option->criteria);

  /* Initialize probability */
  Init_frequency(table, f_new, option->init_type);

  /* Calculate initial likelihood */
  likeli_old = Total_likelihood(table, f_new);
  if (message) {
    printf("Initial log10 likelihood                            = %13.6e\n",likeli_old);
  }

  /* EM iteration loop */
  change = 100.0;
  loop   = 0;
  while (change > criteria) {
    loop++;
    f_tmp = f_new;
    f_new = f_old;
    f_old = f_tmp;

    /* Calculate frequency of all subject */
    Pop_frequency(table, f_old, f_new, f_subject);

    /* Calculate new likelihood */
    likelihood = Total_likelihood(table, f_new);
    change = likelihood - likeli_old;

    if (message && option->message) {
      /* Frequency change : not necessary */
      delta_f_max = 0.0;
      delta_f_sum = 0.0;
      for (h=0; h<table->number_of_haplo; h++) {
	delta_f = *(f_new + h) - *(f_old + h);
	if (delta_f < 0.0) {
	  delta_f = -delta_f;
	}
	if (delta_f > delta_f_max) {
	  delta_f_max = delta_f;
	}
	delta_f_sum = delta_f_sum + delta_f;
      }

      printf("Loop %5d : new log10 likelihood       = %10.3e\n",loop,likelihood);
      printf("             old log10 likelihood       = %10.3e\n",likeli_old);
      printf("             log10 difference (new-old) = %10.3e (criteria %e)\n",change,criteria);
      printf("             maximum frequency change   = %10.3e\n",delta_f_max);
      printf("             sum of frequency change    = %10.3e\n",delta_f_sum);
    }

    likeli_old = likelihood;
    if (change < 0.0) {
      change = -change;
    }
  }

  if (message) {
    printf("Number of iterations to get results                 =  %d\n",loop);
  }
  *f1 = f_old;
  *f2 = f_new;

  return likelihood;
}

int            Init_frequency(Estimate_Set *table,
			      double       *f_init,
			      int           init_type)
{
  /* Local variables */
  haplotype  h, haplo;
  int        l, a;
  double     f, allele_p;

  /* Select initialization type */
  switch (init_type) {
  case 1: /* Equilibrium */
    for (h=0; h<table->number_of_haplo; h++) {
      f     = 1.0;
      haplo = *(table->haplo + h);
      for (l=0; l<table->select_loci; l++) {
	a        = (haplo >> (table->select_loci-l-1)) & 01;
	if (a) {
	  allele_p = 1.0 - *(table->f_allele + l);
	}
	else {
	  allele_p = *(table->f_allele + l);
	}
	f = f * allele_p;
      }
      *(f_init+h) = f;
    }
    break;

  case 2: /* Random */
    f = 0.0;
    for (h=0; h<table->number_of_haplo; h++) {
      *(f_init+h) = random_number();
      f = f + *(f_init+h);
    }
    for (h=0; h<table->number_of_haplo; h++) {
      *(f_init+h) = *(f_init+h) / f;
    }
    break;

  case 0: /* Uniform */
  default: 
    f = 1.0/(double)table->number_of_haplo;
    for (h=0; h<table->number_of_haplo; h++) {
      *(f_init + h) = f;
    }
    break;
  }

  return 0;
}

double           Total_likelihood(Estimate_Set *table,
				  double       *f)
{
  /* Local variables */
  double         likelihood, l_subject;
  int            s, c;
  Diplotype     *diplo;

  /* Initialize likelihood */
  likelihood = 0.0;

  /* Loop for genotype set */
  for (s=0; s<table->number_of_set; s++) {

    /* Loop for haplotype combination */
    l_subject = 0.0;
    diplo = (table->geno_info + s)->combi;
    for (c=0; c<(table->geno_info + s)->number_of_combi; c++) {
      if (diplo->diplotype[0] == diplo->diplotype[1]) {
	l_subject += *(f + diplo->diplotype[0]) * *(f + diplo->diplotype[0]);
      }
      else {
	l_subject += *(f + diplo->diplotype[0]) * *(f + diplo->diplotype[1]) * 2.0;
      }
      diplo++;
    }

    likelihood += log10(l_subject) * (table->geno_info + s)->weight;
  }

  return likelihood;
}

double           Subject_likelihood(Estimate_Set *table,
				    double       *f,
				    double       *l_subject)
{
  /* Local variables */
  double         likelihood;
  int            s, c;
  Diplotype     *diplo;

  /* Initialize likelihood */
  likelihood = 0.0;

  /* Loop for genotype set */
  for (s=0; s<table->number_of_set; s++) {

    /* Loop for haplotype combination */
    *(l_subject+s) = 0.0;
    diplo = (table->geno_info + s)->combi;
    for (c=0; c<(table->geno_info + s)->number_of_combi; c++) {
      if (diplo->diplotype[0] == diplo->diplotype[1]) {
	*(l_subject+s) += *(f + diplo->diplotype[0]) * *(f + diplo->diplotype[0]);
      }
      else {
	*(l_subject+s) += *(f + diplo->diplotype[0]) * *(f + diplo->diplotype[1]) * 2.0;
      }
      diplo++;
    }

    *(l_subject+s) = log10(*(l_subject+s));
    likelihood += *(l_subject+s) * (table->geno_info + s)->weight;
  }

  return likelihood;
}

double         Pop_frequency (Estimate_Set *table,
			      double       *f,
			      double       *f_new,
			      double       *f_subject)
{
  /* Local variables */
  int            s, c;
  haplotype      h;
  Diplotype     *diplo;
  double         norm;

  /* Initialize frequency table */
  for (h=0; h<table->number_of_haplo; h++) {
    *(f_new + h) = 0.0;
  }

  /* Loop for genotype set */
  for (s=0; s<table->number_of_set; s++) {
    for (h=0; h<table->number_of_haplo; h++) {
      *(f_subject + h)   = 0.0;
    }

    /* Loop for haplotype combination */
    diplo = (table->geno_info + s)->combi;
    for (c=0; c<(table->geno_info + s)->number_of_combi; c++) {
      *(f_subject + diplo->diplotype[0]) += *(f + diplo->diplotype[0]) * *(f + diplo->diplotype[1]);
      *(f_subject + diplo->diplotype[1]) += *(f + diplo->diplotype[0]) * *(f + diplo->diplotype[1]);
      diplo++;
    }

    norm = 0.0;
    for (h=0; h<table->number_of_haplo; h++) {
      norm += *(f_subject + h);
    }

    for (h=0; h<table->number_of_haplo; h++) {
      *(f_new     + h) += *(f_subject + h)/norm*(table->geno_info + s)->weight;
    }    
  }

  /*norm = 0.0;*/
  for (h=0; h<table->number_of_haplo; h++) {
    *(f_new + h) = *(f_new + h)*table->normal;
    /*norm += *(f_new + h);*/
  }

  return norm;
}

double         diplotype_probability(int        number_of_combi,
				     Diplotype *combi,
				     double    *f,
				     double    *p)
{
  /* Local variables */
  int          c;
  double       normal;

  /* Loop for combination */
  normal = 0.0;
  for (c=0; c<number_of_combi; c++) {
    *(p + c) = *(f + (combi+c)->diplotype[0]) * *(f + (combi+c)->diplotype[1]);
    normal  += *(p + c);
  }

  for (c=0; c<number_of_combi; c++) {
    *(p + c) = *(p + c)/normal;
  }

  return normal;
}
