#include "haplotypetest.h"

/* Utility function */
extern void          *mem_alloc(size_t, const char *);
extern void           hpsort_reverse_index_of_double_array(int, double *, unsigned int *);

/* Error handler */
extern int            error_check();
extern int            error_message(char *);
extern int            warning_message(char *);

/* Problem definition for EM */
extern int           *filter_loci(Genotype_Data *, Option_Data *, int *);
extern Estimate_Set  *select_window(int, int *, Genotype_Data *, Option_Data *);
extern void           free_estimate_set(Estimate_Set *);

/* Haplotype frequency inference */
extern int            estimate_haplotype(Genotype_Data *, Estimate_Set *, Option_Data *);

/* Incomplete haplotype */
extern void           save_non_zero_haplotype_set(Estimate_Set *, Option_Data *);
extern Haplotype_Set *set_haplotype_set(Genotype_Data *, Estimate_Set *, Option_Data *);
extern void           free_haplotype_set(Haplotype_Set *);

/* Association test */
extern int            estimate_haplotype_and_penetrance(Genotype_Data *, Estimate_Set *, Haplotype_Set *, Option_Data *, Window_Result *);
extern int            estimate_haplotype_and_QTL(Genotype_Data *, Estimate_Set *, Haplotype_Set *, Option_Data *, Window_Result *);

/* Print result */
extern int            print_inheritance_model(int);

/* Define functions */
int            sliding_window_test(Genotype_Data *, Option_Data *);
Window_Result *alloc_window_result(int, int);
void           free_window_result(Window_Result *);
int            print_window_test_with_penetrance(Window_Result *, int, int, int);
int            print_window_test_with_QTL(Window_Result *, int, int, int);

int            sliding_window_test(Genotype_Data *genotype,
				   Option_Data   *option)
{
  /* Local variables */
  int             number_of_all_loci, *all_loci_list;
  int             number_of_windows;
  int             l, window;
  int             model_type;
  Window_Result  *test_result, *t_ptr;

  /* Data structure for test */
  Estimate_Set   *table_for_hap, *table_for_test;
  Haplotype_Set  *set_of_haplotype;

  /* Reset output control flags */
  option->msg_flag      = 0;

  /* Get number of non filtered loci */
  all_loci_list = filter_loci(genotype, option, &number_of_all_loci);
#ifdef DEBUG_WINDOW
  printf("# number of all loci : %d\n# loci :",number_of_all_loci);
  for (l=0; l<number_of_all_loci; l++) {
    printf(" %d",*(all_loci_list+l)+1);
  }
  printf("\n");
#endif

  /* Get number of windows */
  number_of_windows = number_of_all_loci - option->window_size + 1;
  if (number_of_windows < 1) {
    printf("\nWindow size %d is more than number of loci %d. Window size is reset to %d.\n",
	   option->window_size,number_of_all_loci,number_of_all_loci);
    number_of_windows = 1;
    option->window_size = number_of_all_loci;
  }
  printf("\nNumber of windows is %6d\n",number_of_windows);

  /* Allocate sliding window test result array */
  test_result = alloc_window_result(number_of_windows * option->number_of_model, option->window_size);
  t_ptr       = test_result;

  /* Sliding window test */
  for (window=0; window<number_of_windows; window++) {
    /* Select loci */
    table_for_hap  = select_window(window, all_loci_list, genotype, option);
    table_for_test = select_window(window, all_loci_list, genotype, option);

    printf("Window %5d : Loci number =",window+1);
    for (l=0; l<table_for_hap->select_loci; l++) {
      printf(" %d",*(table_for_hap->loci + l)+1);
    }
    printf("\n");

    /* Haplotype inference by EM */
    estimate_haplotype(genotype,table_for_hap,option);

    /* Save nonzero haplotypes */
    save_non_zero_haplotype_set(table_for_hap,option);

    /* Get incomplete haplotype */
    set_of_haplotype = set_haplotype_set(genotype,table_for_hap,option);

    /* Haplotype inference by EM */
    if (option->penetrance) {
      estimate_haplotype_and_penetrance(genotype,table_for_test,set_of_haplotype,option,t_ptr);
    }
      /* Haplotype and QTL parameter inference by EM */
    if (option->QTL) {
      estimate_haplotype_and_QTL(genotype,table_for_test,set_of_haplotype,option,t_ptr);
    }

    /* Free haplotype inference table */
    free_haplotype_set(set_of_haplotype);
    free_estimate_set(table_for_test);
    free_estimate_set(table_for_hap);

    t_ptr = t_ptr + option->number_of_model;
  }

  /* Sliding test result */
  t_ptr       = test_result;
  for (model_type=0; model_type<option->number_of_model; model_type++) {
    if (option->penetrance) {
      print_window_test_with_penetrance(t_ptr, number_of_windows, option->number_of_model, option->model_type[model_type]);
    }
    if (option->QTL) {
      print_window_test_with_QTL(t_ptr, number_of_windows, option->number_of_model, option->model_type[model_type]);
    }

    t_ptr++;
  }  

  /* Free work area */
  free_window_result(test_result);
  free(all_loci_list);

  return 0;
}

Window_Result *alloc_window_result(int number_of_windows,
				   int window_size)
{
  Window_Result  *test_result;
  int             window, name_size, size;
  char           *buffer;

  test_result = (Window_Result *)mem_alloc(number_of_windows*sizeof(Window_Result),"Array for sliding test result");

  name_size = CHAR_IN_ALLELE_NAME * window_size + 1;
  size      = name_size * number_of_windows;
  buffer    = (char *)mem_alloc(size*sizeof(char),"Array for haplotype name of sliding test result");
  for (window=0; window<number_of_windows; window++) {
    (test_result + window)->haplotype_name = buffer;
    buffer = buffer + name_size;
  }

  return test_result;
}

void           free_window_result(Window_Result *test_result)
{
  if (test_result->haplotype_name != NULL) {
    free(test_result->haplotype_name);
  }
  free(test_result);

  return;
}

int            print_window_test_with_penetrance(Window_Result *test_result,
						 int            number_of_windows,
						 int            number_of_model,
						 int            model_type)
{
  /* Local variables */
  int               s;
  Window_Result    *subset;

  /* Print headings */
  print_inheritance_model(model_type);
  printf("\nResult of sliding window test with haplotype set\n");
  if (model_type == GENOTYPE_MODEL) {
    printf("  Window No.  statistics     P value      q0      q1      q2  haplotype set\n");
  }
  else {
    printf("  Window No.  statistics     P value      q-      q+  haplotype set\n");
  }

  subset = test_result;
  for (s=0; s<number_of_windows; s++) {
    if (model_type == GENOTYPE_MODEL) {
      printf("%12d %11.4f %11.4e %7.4f %7.4f %7.4f  %s\n",
	     s+1,subset->statistic,subset->p_value,
	     subset->result[0],subset->result[1],subset->result[2],
	     subset->haplotype_name);
    }
    else {
      printf("%12d %11.4f %11.4e %7.4f %7.4f  %s\n",
	     s+1,subset->statistic,subset->p_value,
	     subset->result[0],subset->result[1],
	     subset->haplotype_name);
    }
    subset = subset + number_of_model;
  }

  return 0;
}

int            print_window_test_with_QTL(Window_Result *test_result,
						 int            number_of_windows,
						 int            number_of_model,
						 int            model_type)
{
  /* Local variables */
  int               i, s;
  Window_Result    *subset;

  /* Print headings */
  print_inheritance_model(model_type);
  printf("\nResult of sliding window test with haplotype set\n");
  if (model_type == GENOTYPE_MODEL) {
    printf("  Window No.  statistics     P value         mu0         mu1         mu2          SD  haplotype set\n");
  }
  else {
    printf("  Window No.  statistics     P value         mu-         mu+          SD  haplotype set\n");
  }

  subset = test_result;
  for (s=0; s<number_of_windows; s++) {
    if (model_type == GENOTYPE_MODEL) {
      printf("%12d %11.4f %11.4e %11.4e %11.4e %11.4e %11.4e  %s\n",
	     s+1,subset->statistic,subset->p_value,
	     subset->result[0],subset->result[1],subset->result[2],subset->result[3],
	     subset->haplotype_name);
    }
    else {
      printf("%12d %11.4f %11.4e %11.4e %11.4e %11.4e  %s\n",
	     s+1,(subset+s)->statistic,(subset+s)->p_value,
	     subset->result[0],subset->result[1],subset->result[3],
	     subset->haplotype_name);
    }
    subset = subset + number_of_model;
  }

  return 0;
}
