#include "haplotypetest.h"

/* Utility function */
extern void          *mem_alloc(size_t, const char *);

/* Error handler */
extern int            error_check();
extern int            error_message(char *);
extern int            warning_message(char *);

/* Random generator */
extern double         random_number();

#ifdef DEBUG
extern char          *binary_string(haplotype);
#endif

/* Define functions */
double         EM_iteration_with_penetrance(Estimate_Set *, Option_Data *, double **, double **, double *, int);
double         Total_likelihood_with_penetrance(Estimate_Set *, double *);
double         Subject_likelihood_with_penetrance(Estimate_Set *, double *, double *);
double         Pop_frequency_with_penetrance(Estimate_Set *, double *, double *, double *);
double         diplotype_probability_with_penetrance(int, Diplotype *, DiploClass *, int, Estimate_Set *, double *, double *);

double         dummy_penetrance[3] = {1.0, 1.0, 1.0};

double         EM_iteration_with_penetrance(Estimate_Set  *table,
					    Option_Data   *option,
					    double       **f1, /* f_old */
					    double       **f2, /* f_new */
					    double        *f_subject,
					    int            message)
{
  /* Frequencies */
  double     *f_old, *f_new, *f_tmp;

  /* Penetrance */

  /* EM calculation parameter */
  double      likelihood, likeli_old;
  double      criteria, change;
  double      delta_f, delta_f_max, delta_f_sum;
  int         loop;
  haplotype   h;

  f_old    = *f1;
  f_new    = *f2;
  criteria = log10(option->criteria);

  /* Initialize probability */
  table->p_plus[0]  = option->p_plus[0];
  table->p_plus[1]  = option->p_plus[1];
  table->p_plus[2]  = option->p_plus[2];
  table->p_minus[0] = 1.0 - table->p_plus[0];
  table->p_minus[1] = 1.0 - table->p_plus[1];
  table->p_minus[2] = 1.0 - table->p_plus[2];

  /* Calculate initial likelihood */
  likeli_old = Total_likelihood_with_penetrance(table, f_new);
  if (message) {
    printf("Initial log10 likelihood                            = %13.6e\n",likeli_old);
  }

  /* EM iteration loop */
  change = 100.0;
  loop   = 0;
  while (change > criteria) {
    loop++;
    f_tmp = f_new;
    f_new = f_old;
    f_old = f_tmp;

    /* Calculate frequency of all subject */
    Pop_frequency_with_penetrance(table, f_old, f_new, f_subject);

    /* Calculate new likelihood */
    likelihood = Total_likelihood_with_penetrance(table, f_new);
    change = likelihood - likeli_old;

    if (message && option->message) {
      /* Frequency change : not necessary */
      delta_f_max = 0.0;
      delta_f_sum = 0.0;
      for (h=0; h<table->number_of_haplo; h++) {
	delta_f = *(f_new + h) - *(f_old + h);
	if (delta_f < 0.0) {
	  delta_f = -delta_f;
	}
	if (delta_f > delta_f_max) {
	  delta_f_max = delta_f;
	}
	delta_f_sum = delta_f_sum + delta_f;
      }

      printf("Loop %5d : new log10 likelihood       = %10.3e\n",loop,likelihood);
      printf("             old log10 likelihood       = %10.3e\n",likeli_old);
      printf("             log10 difference (new-old) = %10.3e (criteria %e)\n",change,criteria);
      printf("             maximum frequency change   = %10.3e\n",delta_f_max);
      printf("             sum of frequency change    = %10.3e\n",delta_f_sum);
    }

    likeli_old = likelihood;
    if (change < 0.0) {
      change = -change;
    }
  }

  if (message) {
    printf("Number of iterations to get results                 =  %d\n",loop);
  }
  *f1 = f_old;
  *f2 = f_new;

  return likelihood;
}

double           Total_likelihood_with_penetrance(Estimate_Set *table,
						  double       *f)
{
  /* Local variables */
  double         likelihood, l_subject;
  int            s, c, p;
  Diplotype     *diplo;
  DiploClass    *class;
  double        *penetrance;

  /* Initialize likelihood */
  likelihood = 0.0;

  /* Loop for genotype set */
  for (s=0; s<table->number_of_set; s++) {
    if      ((table->geno_info + s)->phenotype == 1) {
      penetrance = table->p_plus;
    }
    else if ((table->geno_info + s)->phenotype == 0) {
      penetrance = table->p_minus;
    }
    else {
      penetrance = dummy_penetrance;
    }

    /* Loop for haplotype combination */
    l_subject = 0.0;
    diplo = (table->geno_info + s)->combi;
    class = (table->geno_info + s)->class;
    for (c=0; c<(table->geno_info + s)->number_of_combi; c++) {
      p = (int)class->p_flag;
      if (diplo->diplotype[0] == diplo->diplotype[1]) {
	l_subject += *(f + diplo->diplotype[0]) * *(f + diplo->diplotype[0]) * *(penetrance + p);
      }
      else {
	l_subject += *(f + diplo->diplotype[0]) * *(f + diplo->diplotype[1]) * *(penetrance + p) * 2.0;
      }
      diplo++; class++;
    }

    likelihood += log10(l_subject) * (table->geno_info + s)->weight;
  }

  return likelihood;
}

double           Subject_likelihood_with_penetrance(Estimate_Set *table,
						    double       *f,
						    double       *l_subject)
{
  /* Local variables */
  double         likelihood;
  int            s, c, p;
  Diplotype     *diplo;
  DiploClass    *class;
  double        *penetrance;

  /* Initialize likelihood */
  likelihood = 0.0;

  /* Loop for genotype set */
  for (s=0; s<table->number_of_set; s++) {
    if      ((table->geno_info + s)->phenotype == 1) {
      penetrance = table->p_plus;
    }
    else if ((table->geno_info + s)->phenotype == 0) {
      penetrance = table->p_minus;
    }
    else {
      penetrance = dummy_penetrance;
    }

    /* Loop for haplotype combination */
    *(l_subject+s) = 0.0;
    diplo = (table->geno_info + s)->combi;
    class = (table->geno_info + s)->class;
    for (c=0; c<(table->geno_info + s)->number_of_combi; c++) {
      p = (int)class->p_flag;
      if (diplo->diplotype[0] == diplo->diplotype[1]) {
	*(l_subject+s) += *(f + diplo->diplotype[0]) * *(f + diplo->diplotype[0]) * *(penetrance + p);
      }
      else {
	*(l_subject+s) += *(f + diplo->diplotype[0]) * *(f + diplo->diplotype[1]) * *(penetrance + p) * 2.0;
      }
      diplo++; class++;
    }

    *(l_subject+s) = log10(*(l_subject+s));
    likelihood += *(l_subject+s) * (table->geno_info + s)->weight;
  }

  return likelihood;
}

double         Pop_frequency_with_penetrance(Estimate_Set *table,
					     double       *f,
					     double       *f_new,
					     double       *f_subject)
{
  /* Local variables */
  int            s, c, p, d;
  haplotype      h;
  Diplotype     *diplo;
  DiploClass    *class;
  double        *penetrance;
  double         norm;
  double         count[3][3], count_s[3];
  double         p_plus[3];

  /* Initialize frequency table */
  for (h=0; h<table->number_of_haplo; h++) {
    *(f_new + h) = 0.0;
  }

  /* Initialize counter for new penetrance */
  count[0][0] = 0.0;
  count[0][1] = 0.0;
  count[1][0] = 0.0;
  count[1][1] = 0.0;
  count[2][0] = 0.0;
  count[2][1] = 0.0;

  /* Loop for genotype set */
  for (s=0; s<table->number_of_set; s++) {
    for (h=0; h<table->number_of_haplo; h++) {
      *(f_subject + h)   = 0.0;
    }

    if      ((table->geno_info + s)->phenotype == 1) {
      penetrance = table->p_plus;
      d = 1;
    }
    else if ((table->geno_info + s)->phenotype == 0) {
      penetrance = table->p_minus;
      d = 0;
    }
    else {
      penetrance = dummy_penetrance;
      d = 2;
    }
    count_s[0] = 0.0;
    count_s[1] = 0.0;
    count_s[2] = 0.0;

    /* Loop for haplotype combination */
    diplo = (table->geno_info + s)->combi;
    class = (table->geno_info + s)->class;
    for (c=0; c<(table->geno_info + s)->number_of_combi; c++) {
      p = (int)class->p_flag;
      *(f_subject + diplo->diplotype[0]) += *(f + diplo->diplotype[0]) * *(f + diplo->diplotype[1]) * *(penetrance + p);
      *(f_subject + diplo->diplotype[1]) += *(f + diplo->diplotype[0]) * *(f + diplo->diplotype[1]) * *(penetrance + p);
      count_s[p]                         += *(f + diplo->diplotype[0]) * *(f + diplo->diplotype[1]) * *(penetrance + p);
      diplo++; class++;
    }

    norm = 0.0;
    for (h=0; h<table->number_of_haplo; h++) {
      norm += *(f_subject + h);
    }

    for (h=0; h<table->number_of_haplo; h++) {
      *(f_new     + h)               += *(f_subject + h)/norm*(table->geno_info + s)->weight;
    }

    norm = count_s[0] + count_s[1] + count_s[2];
    count[0][d] += count_s[0]/norm*(table->geno_info + s)->weight;
    count[1][d] += count_s[1]/norm*(table->geno_info + s)->weight;
    count[2][d] += count_s[2]/norm*(table->geno_info + s)->weight;
  }

  for (s=0; s<3; s++) {
    for (c=0; c<2; c++) {
      table->contingency[s][c] = count[s][c];
    }

    count[s][0] = count[s][0]+count[s][1];
    if (count[s][0] != 0.0) {
      p_plus[s] = count[s][1] / count[s][0];
    }
    else {
      p_plus[s] = 0.0;
    }
  }

#ifdef DEBUG_PENETRANCE
  printf("# N_b+x = %8.2f N_b+ = %8.2f p+old = %8.2f p+new = %8.2f\n",count[1][1],count[1][0],table->p_plus[1],p_plus[1]);
  printf("# N_b-x = %8.2f N_b- = %8.2f p-old = %8.2f p-new = %8.2f\n",count[0][1],count[0][0],table->p_plus[0],p_plue[0]);
#endif

  table->p_plus[0]  = p_plus[0];
  table->p_plus[1]  = p_plus[1];
  table->p_plus[2]  = p_plus[2];
  table->p_minus[0] = 1.0 - table->p_plus[0];
  table->p_minus[1] = 1.0 - table->p_plus[1];
  table->p_minus[2] = 1.0 - table->p_plus[2];

  /*norm = 0.0;*/
  for (h=0; h<table->number_of_haplo; h++) {
    *(f_new + h) = *(f_new + h)*table->normal;
    /*norm += *(f_new + h);*/
  }

  return norm;
}

double         diplotype_probability_with_penetrance(int           number_of_combi,
						     Diplotype    *combi,
						     DiploClass   *class,
						     int           phenotype,
						     Estimate_Set *table,
						     double       *f,
						     double       *p_subj)
{
  /* Local variables */
  int          c, p;
  double      *penetrance;
  double       normal;

  /* Loop for combination */
  if      (phenotype == 1) {
    penetrance = table->p_plus;
  }
  else if (phenotype == 0) {
    penetrance = table->p_minus;
  }
  else {
    penetrance = dummy_penetrance;
  }

  normal = 0.0;
  for (c=0; c<number_of_combi; c++) {
    p = (int)(class + c)->p_flag;
    *(p_subj + c) = *(f + (combi+c)->diplotype[0]) * *(f + (combi+c)->diplotype[1]) * *(penetrance + p);
    normal       += *(p_subj + c);
  }

  for (c=0; c<number_of_combi; c++) {
    *(p_subj + c) = *(p_subj + c)/normal;
  }

  return normal;
}
