#include "haplotypetest.h"

/* Utility function */
extern void          *mem_alloc(size_t, const char *);

/* Error handler */
extern int            error_check();
extern int            error_message(char *);
extern int            warning_message(char *);

/* Random generator */
extern double         random_number();

/* Frequency initialization */
extern int            Init_frequency(Estimate_Set *, double *, int);

#ifdef DEBUG
extern char          *binary_string(haplotype);
#endif

/* Define functions */
double         EM_iteration_with_QTL(Estimate_Set *, Option_Data *, Genotype_Data *, double **, double **, double *, int);
double         Total_likelihood_with_QTL(Estimate_Set *, Genotype_Data *, double *);
double         Subject_likelihood_with_QTL(Estimate_Set *, Genotype_Data *, double *, double *);
static double         Pop_frequency_with_QTL(Estimate_Set *, Genotype_Data *, double *, double *, double *);
static int            init_QTL_parameter(Estimate_Set *, Genotype_Data *);
static double         normal_distribution(double, double, double);
double         diplotype_probability_with_QTL(int, Diplotype *, DiploClass *, double, Estimate_Set *, double *, double *);

double         EM_iteration_with_QTL(Estimate_Set  *table,
				     Option_Data   *option,
				     Genotype_Data *genotype,
				     double       **f1, /* f_old */
				     double       **f2, /* f_new */
				     double        *f_subject,
				     int            message)
{
  /* Frequencies */
  double     *f_old, *f_new, *f_tmp;

  /* QTL */
  double      mu_old[3], mu_new[3], SD_new[3], SD_old[3];

  /* EM calculation parameter */
  double      likelihood, likeli_old;
  double      criteria, change;
  double      delta_f, delta_f_max, delta_f_sum;
  int         loop;
  haplotype   h;

  f_old    = *f1;
  f_new    = *f2;
  criteria = log10(option->criteria);

  /* Initialize distribution parameters */
  init_QTL_parameter(table, genotype);
  mu_new[0]    = table->mu[0];
  mu_new[1]    = table->mu[1];
  mu_new[2]    = table->mu[2];
  SD_new[0]    = table->SD_c[0];
  SD_new[1]    = table->SD_c[1];
  SD_new[2]    = table->SD_c[2];

  /* Calculate initial likelihood */
  likeli_old = Total_likelihood_with_QTL(table, genotype, f_new);
  if (message) {
    printf("Initial log10 likelihood                            = %13.6e\n",likeli_old);
  }

  /* EM iteration loop */
  change = 100.0;
  loop   = 0;
  while (change > criteria) {
    loop++;
    f_tmp = f_new;
    f_new = f_old;
    f_old = f_tmp;

    mu_old[0] = mu_new[0];
    mu_old[1] = mu_new[1];
    mu_old[2] = mu_new[2];
    SD_old[0] = SD_new[0];
    SD_old[1] = SD_new[1];
    SD_old[2] = SD_new[2];

    /* Calculate frequency of all subject */
    Pop_frequency_with_QTL(table, genotype, f_old, f_new, f_subject);

    mu_new[0]    = table->mu[0];
    mu_new[1]    = table->mu[1];
    mu_new[2]    = table->mu[2];
    SD_new[0]    = table->SD_c[0];
    SD_new[1]    = table->SD_c[1];
    SD_new[2]    = table->SD_c[2];
    /* Calculate new likelihood */
    likelihood = Total_likelihood_with_QTL(table, genotype, f_new);
    change = likelihood - likeli_old;

    if (message && option->message) {
      /* Frequency change : not necessary */
      delta_f_max = 0.0;
      delta_f_sum = 0.0;
      for (h=0; h<table->number_of_haplo; h++) {
	delta_f = *(f_new + h) - *(f_old + h);
	if (delta_f < 0.0) {
	  delta_f = -delta_f;
	}
	if (delta_f > delta_f_max) {
	  delta_f_max = delta_f;
	}
	delta_f_sum = delta_f_sum + delta_f;
      }

      printf("Loop %5d : new log10 likelihood       = %10.3e\n",loop,likelihood);
      printf("             old log10 likelihood       = %10.3e\n",likeli_old);
      printf("             log10 difference (new-old) = %10.3e (criteria %e)\n",change,criteria);
      printf("             maximum frequency change   = %10.3e\n",delta_f_max);
      printf("             sum of frequency change    = %10.3e\n",delta_f_sum);
    }

    likeli_old = likelihood;
    if (change < 0.0) {
      change = -change;
    }
  }

  if (message) {
    printf("Number of iterations to get results                 =  %d\n",loop);
  }
  *f1 = f_old;
  *f2 = f_new;

  return likelihood;
}

double           Total_likelihood_with_QTL(Estimate_Set  *table,
					   Genotype_Data *genotype,
					   double        *f)
{
  /* Local variables */
  double         likelihood, l_subject;
  int            s, c, p, i;
  Diplotype     *diplo;
  DiploClass    *class;
  double        *penetrance;
  double         density[3];

  /* Initialize likelihood */
  likelihood = 0.0;

  /* Loop for subject */
  for (i=0; i<genotype->Number_of_Subjects; i++) {
    /* Check if phenotype is entered */
    if ((genotype->Subjects + i)->phenotype >= 0) {
      /* Get genotype index */
      s = (genotype->Subjects + i)->packed_offset;

      if (s >= 0) {
	density[0] = normal_distribution((genotype->Subjects + i)->QTL, table->mu[0], table->SD_c[0]);
	density[1] = normal_distribution((genotype->Subjects + i)->QTL, table->mu[1], table->SD_c[1]);
	if (table->model_type == GENOTYPE_MODEL) {
	  density[2] = normal_distribution((genotype->Subjects + i)->QTL, table->mu[2], table->SD_c[2]);
	}

	/* Loop for haplotype combination */
	l_subject = 0.0;
	diplo = (table->geno_info + s)->combi;
	class = (table->geno_info + s)->class;
	for (c=0; c<(table->geno_info + s)->number_of_combi; c++) {
	  p = (int)class->p_flag;

	  /* l_subject changed */
	  if (diplo->diplotype[0] == diplo->diplotype[1]) {
	    l_subject += *(f + diplo->diplotype[0]) * *(f + diplo->diplotype[0]) * density[p];
	  }
	  else {
	    l_subject += *(f + diplo->diplotype[0]) * *(f + diplo->diplotype[1]) * density[p] * 2.0;
	  }

	  diplo++; class++;
	}
	/* End of loop for haplotype combination */

	likelihood += log10(l_subject);
      }
    }
  }
  /* End of loop for subject */

  /* Process phenotype-unknown subjects */
  if (genotype->Number_of_Unknown > 0) {
    /* Loop for genotype set */
    for (s=0; s<table->number_of_set; s++) {
      if ((table->geno_info + s)->phenotype == -1) {

	/* Loop for haplotype combination */
	l_subject = 0.0;
	diplo = (table->geno_info + s)->combi;
	for (c=0; c<(table->geno_info + s)->number_of_combi; c++) {
	  if (diplo->diplotype[0] == diplo->diplotype[1]) {
	    l_subject += *(f + diplo->diplotype[0]) * *(f + diplo->diplotype[0]);
	  }
	  else {
	    l_subject += *(f + diplo->diplotype[0]) * *(f + diplo->diplotype[1]) * 2.0;
	  }
	  diplo++;
	}

	likelihood += log10(l_subject) * (table->geno_info + s)->weight;
      }
    }
    /* End of loop for genotype set */
  }

  return likelihood;
}

double           Subject_likelihood_with_QTL(Estimate_Set  *table,
					     Genotype_Data *genotype,
					     double        *f,
					     double        *l_subject)
{
  /* Local variables */
  double         likelihood, l_subject_u;
  int            s, c, p, i;
  Diplotype     *diplo;
  DiploClass    *class;
  double        *penetrance;
  double         density[3];

  /* Initialize likelihood */
  likelihood = 0.0;

  /* Loop for subject */
  for (i=0; i<genotype->Number_of_Subjects; i++) {
    /* Check if phenotype is entered */
    if ((genotype->Subjects + i)->phenotype >= 0) {
      /* Get genotype index */
      s = (genotype->Subjects + i)->packed_offset;

      if (s >= 0) {
	density[0] = normal_distribution((genotype->Subjects + i)->QTL, table->mu[0], table->SD_c[0]);
	density[1] = normal_distribution((genotype->Subjects + i)->QTL, table->mu[1], table->SD_c[1]);
	if (table->model_type == GENOTYPE_MODEL) {
	  density[2] = normal_distribution((genotype->Subjects + i)->QTL, table->mu[2], table->SD_c[2]);
	}

	/* Loop for haplotype combination */
	*(l_subject+i) = 0.0;
	diplo = (table->geno_info + s)->combi;
	class = (table->geno_info + s)->class;
	for (c=0; c<(table->geno_info + s)->number_of_combi; c++) {
	  p = (int)class->p_flag;

	  /* l_subject changed */
	  if (diplo->diplotype[0] == diplo->diplotype[1]) {
	    *(l_subject+i) += *(f + diplo->diplotype[0]) * *(f + diplo->diplotype[0]) * density[p];
	  }
	  else {
	    *(l_subject+i) += *(f + diplo->diplotype[0]) * *(f + diplo->diplotype[1]) * density[p] * 2.0;
	  }

	  diplo++; class++;
	}
	/* End of loop for haplotype combination */

	*(l_subject+i) = log10(*(l_subject+i));
	likelihood += *(l_subject+i);
      }
      else {
	*(l_subject+i) = 0.0;
      }
    }
  }
  /* End of loop for subject */

  /* Process phenotype-unknown subjects */
  if (genotype->Number_of_Unknown > 0) {
    /* Loop for genotype set */
    for (s=0; s<table->number_of_set; s++) {
      if ((table->geno_info + s)->phenotype == -1) {

	/* Loop for haplotype combination */
	l_subject_u = 0.0;
	diplo = (table->geno_info + s)->combi;
	for (c=0; c<(table->geno_info + s)->number_of_combi; c++) {
	  if (diplo->diplotype[0] == diplo->diplotype[1]) {
	    l_subject_u += *(f + diplo->diplotype[0]) * *(f + diplo->diplotype[0]);
	  }
	  else {
	    l_subject_u += *(f + diplo->diplotype[0]) * *(f + diplo->diplotype[1]) * 2.0;
	  }
	  diplo++;
	}

	likelihood += log10(l_subject_u) * (table->geno_info + s)->weight;
      }
    }
    /* End of loop for genotype set */
  }

  return likelihood;
}

double         Pop_frequency_with_QTL(Estimate_Set  *table,
				      Genotype_Data *genotype,
				      double        *f,
				      double        *f_new,
				      double        *f_subject)
{
  /* Local variables */
  int            s, c, p, i;
  haplotype      h;
  Diplotype     *diplo;
  DiploClass    *class;
  double         norm;

  double         density[3], diplo_prob[3];
  double         indiv_denomi[3], indiv_numera[3];
  double         overall_denomi[3], overall_mu_numera[3], overall_SD_numera[3];
  double         a_i[3];

  /* Initialize frequency table */
  for (h=0; h<table->number_of_haplo; h++) {
    *(f_new + h) = 0.0;
  }

  /* Initialize distribution parameters etc. */
  for (p=0; p<3; p++) {
    overall_denomi[p]    = 0.0;
    overall_mu_numera[p] = 0.0;
    overall_SD_numera[p] = 0.0;
  }

  /* Loop for individual */
  for (i=0;i<genotype->Number_of_Subjects;i++){
    /* Check if phenotype is entered */
    if ((genotype->Subjects + i)->phenotype >= 0) {
      /* Get genotype index */
      s = (genotype->Subjects + i)->packed_offset;

      if (s >= 0) {
	for (h=0; h<table->number_of_haplo; h++) {
	  *(f_subject + h)   = 0.0;
	}

	for (p=0; p<3; p++) {
	  indiv_denomi[p] = 0.0;
	  indiv_numera[p] = 0.0;
	}

	density[0] = normal_distribution((genotype->Subjects + i)->QTL, table->mu[0], table->SD_c[0]);
	density[1] = normal_distribution((genotype->Subjects + i)->QTL, table->mu[1],  table->SD_c[1]);
	if (table->model_type == GENOTYPE_MODEL) {
	  density[2] = normal_distribution((genotype->Subjects + i)->QTL, table->mu[2], table->SD_c[2]);
	}

	/* Loop for haplotype combination */
	diplo = (table->geno_info + s)->combi;
	class = (table->geno_info + s)->class;
	for (c=0; c<(table->geno_info + s)->number_of_combi; c++) {
	  p = (int)class->p_flag;

	  diplo_prob[0]   = *(f + diplo->diplotype[0]) * *(f + diplo->diplotype[1]) * density[0];
	  diplo_prob[1]   = *(f + diplo->diplotype[0]) * *(f + diplo->diplotype[1]) * density[1];
	  diplo_prob[2]   = *(f + diplo->diplotype[0]) * *(f + diplo->diplotype[1]) * density[2];

	  *(f_subject + diplo->diplotype[0]) += diplo_prob[p];
	  *(f_subject + diplo->diplotype[1]) += diplo_prob[p];

	  indiv_denomi[0] += diplo_prob[0];
	  indiv_denomi[1] += diplo_prob[1];
	  indiv_denomi[2] += diplo_prob[2];
	  indiv_numera[p] += diplo_prob[p];

	  diplo++; class++;
	}
	/* End of loop for haplotype combination */

      
	for (p=0; p<3; p++) {
	  if (indiv_denomi[p] != 0.0) {
	    a_i[p]                = indiv_numera[p] / indiv_denomi[p];
	    overall_denomi[p]    += a_i[p];
	    overall_mu_numera[p] += (genotype->Subjects+i)->QTL * a_i[p];
	    overall_SD_numera[p] += ((genotype->Subjects+i)->QTL - table->mu[p]) * ((genotype->Subjects+i)->QTL - table->mu[p]) * a_i[p];
	  }
	}

	norm = 0.0;
	for (h=0; h<table->number_of_haplo; h++) {
	  norm += *(f_subject + h);
	}

	for (h=0; h<table->number_of_haplo; h++) {
	  *(f_new + h) += *(f_subject + h)/norm;
	}
      }
    }
  }
  /* End of loop for individual */

  /* Process phenotype-unknown subjects */
  if (genotype->Number_of_Unknown > 0) {
    /* Loop for genotype set */
    for (s=0; s<table->number_of_set; s++) {
      if ((table->geno_info + s)->phenotype == -1) {
	for (h=0; h<table->number_of_haplo; h++) {
	  *(f_subject + h)   = 0.0;
	}

	/* Loop for haplotype combination */
	diplo = (table->geno_info + s)->combi;
	for (c=0; c<(table->geno_info + s)->number_of_combi; c++) {
	  *(f_subject + diplo->diplotype[0]) += *(f + diplo->diplotype[0]) * *(f + diplo->diplotype[1]);
	  *(f_subject + diplo->diplotype[1]) += *(f + diplo->diplotype[0]) * *(f + diplo->diplotype[1]);
	  diplo++;
	}

	norm = 0.0;
	for (h=0; h<table->number_of_haplo; h++) {
	  norm += *(f_subject + h);
	}

	for (h=0; h<table->number_of_haplo; h++) {
	  *(f_new     + h) += *(f_subject + h)/norm*(table->geno_info + s)->weight;
	}
      }
    }
    /* End of loop for genotype set */
  }

  /* Calculate distribution parameters */
  for (p=0; p<3; p++) {
    table->contingency[p][0] = overall_denomi[p];

    if (overall_denomi[p] != 0.0){
      table->mu[p]   = overall_mu_numera[p] / overall_denomi[p];
      table->SD_c[p] = sqrt(overall_SD_numera[p] / overall_denomi[p]);
    }
    else{
      table->mu[p]   = 0.0;
      table->SD_c[p] = 0.0;
    }    
  }
  table->SD       = sqrt((overall_SD_numera[0]+overall_SD_numera[1]+overall_SD_numera[2]) * table->normal_given);

#ifdef DEBUG_QTL
  printf("# %f %f %f %f",overall_denomi[0],overall_denomi[1],overall_denomi[2],
	 (overall_denomi[0] + overall_denomi[1] + overall_denomi[2]));
#endif

  for (h=0; h<table->number_of_haplo; h++) {
    *(f_new + h) = *(f_new + h) * table->normal;
  }

#ifdef DEBUG_QTL
  printf(" %f\n",*f_new);
#endif

  return norm;
}

int            init_QTL_parameter(Estimate_Set  *table,
				  Genotype_Data *genotype)
{
  int i;
  double xsum, x2sum;
  double ans, ans2;

  xsum  = 0.0;
  x2sum = 0.0;
  for(i=0;i<genotype->Number_of_Subjects;i++){
    xsum  += (genotype->Subjects + i)->QTL;
   x2sum += (genotype->Subjects + i)->QTL * (genotype->Subjects + i)->QTL;
   /* x2sum += ((genotype->Subjects + i)->QTL-(xsum/(double)genotype->Number_of_Subjects))*((genotype->Subjects + i)->QTL-(xsum/(double)genotype->Number_of_Subjects));
  */}

  ans             = xsum / (double)genotype->Number_of_Subjects;
  table->mu[0]    = ans;
  table->mu[1]    = ans;
  table->mu[2]    = ans;

 ans2      = (x2sum - genotype->Number_of_Subjects * ans * ans)/((double)genotype->Number_of_Subjects);
  /* ans2      = x2sum/(double)genotype->Number_of_Subjects;*/
  table->SD = sqrt(ans2);
  table->SD_c[0]  = sqrt(ans2);
  table->SD_c[1]  = sqrt(ans2);
  table->SD_c[2]  = sqrt(ans2);

#ifdef DEBUG_QTL2
  printf("# Init_parm:xsum=%8.4f N=%d mu=%8.4f\n",xsum,genotype->Number_of_Subjects,ans);
  printf("# Init_parm:mu=%8.4f std=%8.4f\n",table->mu_plus,table->std);
#endif

  return 0;
}

double         normal_distribution(double  x,  /* QTL value */
				   double  mu, /* mean of the distribution */
				   double  SD) /* st. dev. of the distribution */
{
  /* Probability Density Function for Normal_Distribution */

  double     dist;

  dist = (x - mu) * (x - mu);
  dist = dist / ( 2.0 * SD * SD);
  dist = -1.0 * dist;
  dist = exp(dist);
  dist = dist / (SD * sqrt(2.0 * PI));

  return dist;
}

double         diplotype_probability_with_QTL(int           number_of_combi,
					      Diplotype    *combi,
					      DiploClass   *class,
					      double        QTL,
					      Estimate_Set *table,
					      double       *f,
					      double       *p_subj)
{
  /* Local variables */
  int          c, p;
  double       normal;
  double       density[3];

  /* Loop for combination */
  density[0] = normal_distribution(QTL, table->mu[0], table->SD_c[0]);
  density[1] = normal_distribution(QTL, table->mu[1], table->SD_c[1]);
  if (table->model_type == GENOTYPE_MODEL) {
    density[2] = normal_distribution(QTL, table->mu[2], table->SD_c[2]);
  }

  normal = 0.0;
  for (c=0; c<number_of_combi; c++) {
    p = (int)(class + c)->p_flag;
    *(p_subj + c) = *(f + (combi+c)->diplotype[0]) * *(f + (combi+c)->diplotype[1]) * density[p];
    normal       += *(p_subj + c);
  }

  for (c=0; c<number_of_combi; c++) {
    *(p_subj + c) = *(p_subj + c)/normal;
  }

  return normal;
}
