#include "haplotypetest.h"

/* Utility function */
extern void          *mem_alloc(size_t, const char *);
extern void           hpsort_reverse_index_of_double_array(int, double *, unsigned int *);
extern void           hpsort_index_of_double_array(int, double *, unsigned int *);

/* Error handler */
extern int            error_check();
extern int            error_message(char *);
extern int            warning_message(char *);

/* Problem definition for EM */
extern Estimate_Set  *setup_EM(Genotype_Data *, Estimate_Set *, int);
extern Estimate_Set  *combination(Estimate_Set *);
extern int            all_haplotype(Estimate_Set *);
extern void           free_estimate_set(Estimate_Set *);

/* EM engine */
extern double         EM_iteration(Estimate_Set *, Option_Data *, double **, double **, double *, int);

/* Functions for no LD condition */
extern double         noLD_likelihood(Estimate_Set *, double *, int);
extern int            noLD_frequency(Estimate_Set *, double *, double *);

/* Result print */
extern int            print_allele_table(Estimate_Set *, Genotype_Data *, double *);
extern int            print_haplotype_frequency(Estimate_Set *, double *, Genotype_Data *);
extern int            print_diplotype_profile(Estimate_Set *, Genotype_Data *);
extern int            print_p_value_of_LRC(int, double);

/* Standard error calculation by non-parametric bootstrap */
extern int            bootstrap_for_haplotype(Genotype_Data *, Estimate_Set *, Option_Data *);

/* Outlier detection */
extern double         Subject_likelihood(Estimate_Set *, double *, double *);
extern int            Outlier_by_genotype_set(Estimate_Set *, Option_Data *, Genotype_Data *, double *, double ());

/* Define functions */
int            estimate_haplotype(Genotype_Data *, Estimate_Set *, Option_Data *);
int            count_nonzero_haplotype(int, double *, haplotype *);
int            estimated_allele_frequency(Estimate_Set *, int, haplotype *, double *, double *);

int            estimate_haplotype(Genotype_Data *genotype,
				  Estimate_Set  *table,
				  Option_Data   *option)
{
  /* Local variables */
  int         size, l, nonzero, init_save, count_max, index_max, maximize_f_save;
  haplotype   h, mask;
  double      criteria;

  /* Frequencies */
  double     *f_old, *f_new, *f_subject, *f_noLD, *f_tmp;
  double     *f_allele_save, *f_allele;
  haplotype  *order, *order_tmp;

  /* Parameter for maximum likelihood estimation */
  double      likelihood, likelihood0, LRC, LRC_max;

  double    (*likelihood_function)(Estimate_Set *, double *, double *) = Subject_likelihood;

  /* Set up tables for haplotype inferance */
  table = setup_EM(genotype, table, option->missing_loci);
  maximize_f_save = option->maximize_f;
  if (option->msg_flag) {
    printf("\nNumber of effective subjects to be used             =  %d\n",genotype->Number_of_Effective_Subjects);
    printf("Number of genotype for EM calculation               =  %d\n",table->number_of_set);
  }
  else {
    option->maximize_f = 0;
  }

  /* Make combination table */
  table = combination(table);
  all_haplotype(table);
  if (option->msg_flag) {
    printf("Number of possible haplotypes in the problem        =  %d\n",table->number_of_haplo);
  }

  /* Allocate frequencies and expectation */
  size         = table->number_of_haplo;
  f_old        = (double *)mem_alloc(size*sizeof(double),"Frequency array for old step");
  f_new        = (double *)mem_alloc(size*sizeof(double),"Frequency array for new step");
  f_subject    = (double *)mem_alloc(size*sizeof(double),"Frequency array for each subject");

  if (option->maximize_f) {
    printf("Restart count to search maximum LRC                 =  %d\n",option->maximize);
    printf("\nInformation for first EM iteration of given initialization method as follows\n");
  }

  /* EM iteration */
  likelihood   = EM_iteration(table, option, &f_old, &f_new, f_subject, option->msg_flag);

  /* Sort haplotype frequency */
  table->order   = (haplotype *)mem_alloc(size*sizeof(haplotype),"Array for sort frequency");
  table->f       = f_new;
  hpsort_reverse_index_of_double_array(table->number_of_haplo, table->f, table->order);
  table->nonzero = count_nonzero_haplotype(table->number_of_haplo, table->f, table->order);

  /* Final allele frequency */
  f_allele_save  = (double *)mem_alloc(table->select_loci*sizeof(double),"Allele frequency for selected loci");
  for (l=0; l<table->select_loci; l++) {
    *(f_allele_save + l)   = *(table->f_allele + l);
    *(table->f_allele + l) = 0.0;
  }
  estimated_allele_frequency(table, table->nonzero, table->order, table->f, table->f_allele);
 
  /* No LD likelihood */
  f_noLD = f_old;
  noLD_frequency(table, table->f_allele, f_noLD);
  likelihood0  = noLD_likelihood(table, f_noLD, option->msg_flag);
  LRC_max      = likelihood - likelihood0;
  if (option->msg_flag) {
    printf("Final log10 Likelihood                              = %13.6e\n",likelihood);
    printf("Likelihood Ratio Criteria (LRC)                     = %13.6e\n",LRC_max);
  }

  /* Maximum LRC search */
  if (option->maximize > 0) {
    if (option->maximize_f) {
      /* Output haplotype frequencies and diplotype configuration */
      print_allele_table(table, genotype, f_allele_save);
      print_haplotype_frequency(table, f_noLD, genotype);
    }

    if (option->msg_flag) {
      printf("\nInformation of restart to search maximum LRC\n");
    }
    criteria          = log10(option->criteria)*2.0*fabs(LRC_max);
    count_max         = 0;
    index_max         = 0;
    init_save         = option->init_type;
    option->init_type = 2; /* Random distribution */

    /* Allocate frequencies */
    f_new    = (double *)mem_alloc(size*sizeof(double),"Frequency array for new step");
    f_old    = (double *)mem_alloc(size*sizeof(double),"Frequency array for old step");
    f_allele = (double *)mem_alloc(table->select_loci*sizeof(double),"Allele frequency for selected loci");
    order    = (haplotype *)mem_alloc(size*sizeof(haplotype),"Array for sort frequency");

    /* Loop for LRC search */
    for (l=1; l<=option->maximize; l++){
      if (option->maximize_f) {
	printf("\nRestart No. %d\n",l);
      }

      /* EM iteration */
      likelihood   = EM_iteration(table, option, &f_old, &f_new, f_subject, option->maximize_f);

      /* Sort haplotype frequency */
      hpsort_reverse_index_of_double_array(table->number_of_haplo, f_new, order);
      nonzero = count_nonzero_haplotype(table->number_of_haplo, f_new, order);

      /* Final allele frequency */
      estimated_allele_frequency(table, nonzero, order, f_new, f_allele);

      /* No LD likelihood */
      noLD_frequency(table, f_allele, f_old);
      likelihood0  = noLD_likelihood(table, f_old, option->maximize_f);
      LRC          = likelihood - likelihood0;

      if (option->maximize_f) {
	printf("Final log10 Likelihood                              = %13.6e\n",likelihood);
	printf("Likelihood Ratio Criteria (LRC)                     = %13.6e\n",LRC);
      }

      if      (fabs(LRC - LRC_max) < criteria) {
	count_max++;
      }
      else if (LRC - LRC_max > criteria) {
	count_max       = 0;
	index_max       = l;
	LRC_max         = LRC;

	f_tmp           = table->f;
	table->f        = f_new;
	f_new           = f_tmp;
	f_tmp           = f_noLD;
	f_noLD          = f_old;
	f_old           = f_tmp;
	f_tmp           = table->f_allele;
	table->f_allele = f_allele;
	f_allele        = f_tmp;
	order_tmp       = table->order;
	table->order    = order;
	order           = order_tmp;
	table->nonzero  = nonzero;

	if (option->msg_flag) {
	  printf("Restart No. %5d : LRC maximum                     = %13.6e\n",l,LRC);
	}

	if (option->maximize_f) {
	  /* Output haplotype frequencies and diplotype configuration */
	  print_allele_table(table, genotype, f_allele_save);
	  print_haplotype_frequency(table, f_noLD, genotype);
	}
      }
    }
    if (option->msg_flag) {
      printf("  Final LRC                                         = %13.6e at %d\n",LRC_max,index_max);
      printf("  Same LRC came out %d times\n",count_max);
    }
    
    option->init_type = init_save;
    free(order);
    free(f_allele);
    free(f_old);
    free(f_new);
  }

  /* Output haplotype frequencies and diplotype configuration */
  if (option->msg_flag) {
    print_p_value_of_LRC(table->df, LRC_max);
    print_allele_table(table, genotype, f_allele_save);
    print_haplotype_frequency(table, f_noLD, genotype);
    if (option->outlier) {
      Outlier_by_genotype_set(table, option, genotype, table->f, likelihood_function);
    }
    if (option->diplo_out) {
      print_diplotype_profile(table, genotype);
    }
  }

  option->maximize_f = maximize_f_save;
  free(f_allele_save);
  free(f_subject);
  free(f_noLD);

  /* Bootstrap calculation start */
  if (option->bootstrap > 0) {
    bootstrap_for_haplotype(genotype, table, option);
  }
  /* Bootstrap calculation end */

  return 0;
}

int            count_nonzero_haplotype(int        number_of_haplo,
				       double    *f,
				       haplotype *order)
{
  /* Local variables */
  haplotype  h;
  int        nonzero;

  nonzero = number_of_haplo;
  for (h=0; h<number_of_haplo; h++) {
    if (*(f + *(order + h)) < MIN_FREQUENCY) {
      nonzero = h; break;
    }
  }

  return nonzero;
}

int            estimated_allele_frequency(Estimate_Set *table,
					  int           nonzero,
					  haplotype    *order,
					  double       *f,
					  double       *f_allele)
{
  /* Local variables */
  int        l;
  haplotype  h, mask;

  for (l=0; l<table->select_loci; l++) {
    *(f_allele + l) = 0.0;
    mask = 01 << (table->select_loci - l - 1);
    for (h=0; h<nonzero; h++) {
      if ((*(table->haplo + *(order + h)) & mask) == 0) {
	*(f_allele + l) += *(f + *(order + h));
      }
    }
  }

  return 0;
}
