#include "haplotypetest.h"

/* Utility function */
extern void          *mem_alloc(size_t, const char *);
extern void           hpsort_reverse_index_of_double_array(int, double *, unsigned int *);

/* Error handler */
extern int            error_check();
extern int            error_message(char *);
extern int            warning_message(char *);

/* Problem definition for EM */
extern Estimate_Set  *alloc_estimate_set();
extern int           *filter_loci(Genotype_Data *, Option_Data *, int *);
extern int            set_target_haplotype(Genotype_Data *, Estimate_Set *, Option_Data *);
extern Estimate_Set  *setup_EM(Genotype_Data *, Estimate_Set *, int);
extern Estimate_Set  *combination(Estimate_Set *);
extern void           free_estimate_set(Estimate_Set *);

/* EM engine */
extern double         EM_iteration(Estimate_Set *, Option_Data *, double **, double **, double *, int);

/* Functions for no LD condition */
extern double         noLD_likelihood(Estimate_Set *, double *, int);
extern int            noLD_frequency(Estimate_Set *, double *, double *);

/* Result print */
extern int            count_nonzero_haplotype(int, double *, haplotype *);
extern int            print_allele_table(Estimate_Set *, Genotype_Data *, double *);
extern int            print_haplotype_frequency(Estimate_Set *, double *, Genotype_Data *);

/* Define functions */
int            estimate_LD_measure(Genotype_Data *, Option_Data *);
double         calc_Dprime(Estimate_Set *, double);
double         calc_rsquare(Estimate_Set *, double);
double         calc_Q(Estimate_Set *);
double         calc_Haplotype_Distance(int, double *, double *);
double         calc_Haplotype_Entropy(int, double *);

int            estimate_LD_measure(Genotype_Data *genotype,
				   Option_Data   *option)
{
  /* Local variables */
  haplotype       h, mask;
  int             number_of_loci, *loci_list, size;
  int             l, l1, l2;
  Estimate_Set   *table_pair;
  int             header_flag;

  /* Frequencies */
  double         *f_old, *f_new, *f_subject;
  double         *f_allele;

  /* Parameter for maximum likelihood estimation */
  double      likelihood, likelihood0, LRC;

  /* Linkage disequilibrium measure */
  double      D, Dp, r, Q, d, S;

  /* Memory allocation for the problem definition table */
  size       = MAX_NUMBER_OF_ALLELE*MAX_NUMBER_OF_ALLELE;
  table_pair = alloc_estimate_set();
  table_pair->select_loci = MAX_NUMBER_OF_ALLELE;;
  table_pair->loci        = (int *)      mem_alloc(2   *sizeof(int),      "Loci selection table");
  table_pair->f_allele    = (double *)   mem_alloc(2   *sizeof(double),   "Allele frequency for selected loci");
  table_pair->haplo       = (haplotype *)mem_alloc(size*sizeof(haplotype),"Possible haplotype array");
  table_pair->order       = (haplotype *)mem_alloc(size*sizeof(haplotype),"Array for sort frequency");

  /* Initialize possible haplotype */
  table_pair->number_of_haplo = size;
  for (h=0; h<table_pair->number_of_haplo; h++) {
    *(table_pair->haplo + h) = h;
  }

  /* Allocate frequencies and expectation */
  f_old     = (double *)mem_alloc(size*sizeof(double),"Frequency array for old step");
  f_new     = (double *)mem_alloc(size*sizeof(double),"Frequency array for new step");
  f_subject = (double *)mem_alloc(size*sizeof(double),"Frequency array for each subject");
  f_allele  = (double *)mem_alloc(2   *sizeof(double),"Allele frequency for selected loci");

  /* Get number of non filtered loci */
  loci_list = filter_loci(genotype, option, &number_of_loci);

  if (number_of_loci == 0) {
    warning_message("There are no loci to calculate linkage disequilibrium measure");
    return;
  }

  /* Loop for loci pair */
  header_flag = 1;
  for (l1=0; l1<number_of_loci-1; l1++) {
    for (l2=l1+1; l2<number_of_loci; l2++) {
      *(table_pair->loci)   = *(loci_list + l1);
      *(table_pair->loci+1) = *(loci_list + l2);
      *f_allele             = (genotype->Locus_List + *(loci_list + l1))->Allele_Frequency;
      *(f_allele+1)         = (genotype->Locus_List + *(loci_list + l2))->Allele_Frequency;

      /* Set up tables for haplotype inferance */
      table_pair = setup_EM(genotype, table_pair, 0);

      /* Make combination table */
      table_pair = combination(table_pair);

      /* EM iteration */
      likelihood = EM_iteration(table_pair, option, &f_old, &f_new, f_subject, 0);
      table_pair->f = f_new;

      *(table_pair->f_allele)   = 0.0;
      *(table_pair->f_allele+1) = 0.0;
      for (l=0; l<table_pair->select_loci; l++) {
	mask = 01 << (table_pair->select_loci - l - 1);
	for (h=0; h<table_pair->number_of_haplo; h++) {
	  if ((*(table_pair->haplo + h) & mask) == 0) {
	    *(table_pair->f_allele + l) += *(table_pair->f + h);
	  }
	}
      }

      noLD_frequency(table_pair, table_pair->f_allele, f_old);
      likelihood0  = noLD_likelihood(table_pair, f_old, 0);
      LRC          = likelihood - likelihood0;
      D            = *(table_pair->f) - *f_old;
      Dp           = calc_Dprime (table_pair, D);
      r            = calc_rsquare(table_pair, D);
      Q            = calc_Q      (table_pair);
      d            = calc_Haplotype_Distance(table_pair->number_of_haplo, table_pair->f, f_old);
      S            = calc_Haplotype_Entropy(table_pair->number_of_haplo, table_pair->f);

      if (header_flag) {
	header_flag = 0;
	printf("\nLoci pair linkage disequilibrium measure based on pair iteration\n");
	printf("     i     j       D      D'     r^2       Q     LRC Distance Entropy\n");
      }
      printf(" %5d %5d %7.4f %7.4f %7.4f %7.4f %7.2f %8.4f %7.4f\n",*(loci_list+l1)+1,*(loci_list+l2)+1,D,Dp,r,Q,LRC,d,S);

      if (option->message) {
	/* No LD likelihood */
	printf("Final log10 Likelihood                              = %13.6e\n",likelihood);
	printf("Likelihood Ratio Criteria (LRC)                     = %13.6e\n",LRC);

      /* Output haplotype frequencies and diplotype configuration */
	print_allele_table(table_pair, genotype, f_allele);
	hpsort_reverse_index_of_double_array(table_pair->number_of_haplo, table_pair->f, table_pair->order);
	table_pair->nonzero = count_nonzero_haplotype(table_pair->number_of_haplo, table_pair->f, table_pair->order);
	print_haplotype_frequency(table_pair, f_old, genotype);

	header_flag = 1;
      }

      /* Free loci dependent area */
      free(table_pair->combi);
      free(table_pair->geno_info);
      table_pair->combi     = NULL;
      table_pair->geno_info = NULL;
      table_pair->f         = NULL;
    }
  }

  /* Free work area */
  free(loci_list);
  free(f_allele);
  free(f_subject);
  free(f_new);
  free(f_old);

  /* Free haplotype inference table */
  free_estimate_set(table_pair);

  return 0;
}

double         calc_Dprime(Estimate_Set *table,
			   double        D)
{
  double  d00, d01, d10, d11, Dmax, Dp;

  if (D > 0) {
    d01 =        *(table->f_allele)  * (1.0 - *(table->f_allele + 1));
    d10 = (1.0 - *(table->f_allele)) *        *(table->f_allele + 1);

    if (d01 > d10) {
      Dmax = d10;
    }
    else {
      Dmax = d01;
    }

    Dp = D/Dmax;
  }
  else {
    d00 =        *(table->f_allele)  *        *(table->f_allele + 1);
    d11 = (1.0 - *(table->f_allele)) * (1.0 - *(table->f_allele + 1));

    if (d00 > d11) {
      Dmax = d11;
    }
    else {
      Dmax = d00;
    }

    /*    Dp = D/Dmax * (-1.0);*/
    Dp = D/Dmax;
  }

  return Dp;
}

double         calc_rsquare(Estimate_Set *table,
			    double        D)
{
  double r;

  r = D * D / (        *(table->f_allele) * 
                       *(table->f_allele + 1) *
                (1.0 - *(table->f_allele)) *
		       (1.0 - *(table->f_allele + 1)) );

  return r;
}

double         calc_Q(Estimate_Set *table)
{
  double Q, X, Y;

  X = *(table->f    ) * *(table->f + 3) - *(table->f + 1) * *(table->f + 2);
  Y = *(table->f    ) * *(table->f + 3) + *(table->f + 1) * *(table->f + 2);
  Q = X / Y;

  return Q;
}

double         calc_Haplotype_Entropy(int     count,
				      double *p_haplo)
{
  double entropy;
  int    i;

  entropy = 0.0;
  for (i=0; i<count; i++) {
    if (*p_haplo > 0) {
      entropy = entropy - *p_haplo * log10(*p_haplo);
    }
    p_haplo++;
  }

  return entropy;
}

double         calc_Haplotype_Distance(int     count,
				       double *p_haplo,
				       double *p_noLD)
{
  double distance;
  int    i;

  distance = 0.0;
  for (i=0; i<count; i++) {
    if (*p_haplo > 0) {
      distance = distance + *p_haplo * log10(*p_haplo / *p_noLD);
    }
    p_haplo++; p_noLD++;
  }

  return distance;
}
