#include "haplotypetest.h"

/* Utility function */
extern void          *mem_alloc(size_t, const char *);
extern void          *mem_realloc(void *, size_t, const char *);
extern void           hpsort_reverse_index_of_double_array(int, double *, unsigned int *);
extern void           sort_int_array(int, unsigned int *); 

/* Error handler */
extern int            error_check();
extern int            error_message(char *);
extern int            warning_message(char *);

/* Random generator */
extern double         random_number();

/* Problem definition for EM */
extern Estimate_Set  *alloc_estimate_set();
extern Estimate_Set  *setup_EM(Genotype_Data *, Estimate_Set *, int);
extern Estimate_Set  *combination(Estimate_Set *);
extern int            all_haplotype(Estimate_Set *);
extern int            set_diplotype_class(Estimate_Set *);
extern int            trait_inheritance_model(Estimate_Set *, int, int);
extern int            no_association_model(Estimate_Set *);

/* Frequency initialization */
extern int            Init_frequency(Estimate_Set *, double *, int);

/* EM engine */
extern double         EM_iteration_with_penetrance(Estimate_Set *, Option_Data *, double **, double **, double *, int);
extern double         EM_iteration_with_QTL(Estimate_Set *, Option_Data *, Genotype_Data *, double **, double **, double *, int);

#ifdef DEBUG_PERMUTATION
extern int            dump_estimate_set(Estimate_Set *);
#endif

double         permutation_with_penetrance(Genotype_Data *, Estimate_Set *, Option_Data *, double, int);
double         permutation_with_QTL(Genotype_Data *, Estimate_Set *, Option_Data *, double, int);

double         permutation_with_penetrance(Genotype_Data *genotype,
					   Estimate_Set  *table,
					   Option_Data   *option,
					   double         LRC_orig,
					   int            message)
{
  /* Local variables */
  double       p_value;
  int          s, p, l, count_upper, init_save;
  int         *phenotype, *order;
  double      *random, *f_new, *f_old, *f_subject, *f_0;
  double       likelihood, likelihood0,  LRC_max, criteria;
  double       likelihood_max, likelihood0_max;
  haplotype    h;

  /* Data structure for permutation */
  Genotype_Data *genotype_for_P;
  Estimate_Set  *table_for_P;

  /* Message */
  if (message) {
    printf("\nPermutation test : number of samples = %d\n",
	   option->permutation);
  }

  /* Allocate data structure */
  genotype_for_P           = (Genotype_Data *)mem_alloc(sizeof(Genotype_Data),"Genotype data structure");
  *genotype_for_P          = *genotype;
  genotype_for_P->Subjects = (Subject_Genotype *)mem_alloc(genotype->Number_of_Effective_Subjects*sizeof(Subject_Genotype),"Subject array");

  table_for_P              = alloc_estimate_set();
  *table_for_P             = *table;

#ifdef DEBUG_PERMUTATION
  dump_estimate_set(table_for_P);
#endif

  /* Allocate data area */
  phenotype = (int *)   mem_alloc(genotype->Number_of_Effective_Subjects*sizeof(int),   "Phenotype array for permutation");
  order     = (int *)   mem_alloc(genotype->Number_of_Effective_Subjects*sizeof(int),   "Index array for phenotype");
  random    = (double *)mem_alloc(genotype->Number_of_Effective_Subjects*sizeof(double),"Array to randomize phenotype");
  f_old     = (double *)mem_alloc(table->number_of_haplo*sizeof(double),"Frequency array for old step");
  f_new     = (double *)mem_alloc(table->number_of_haplo*sizeof(double),"Frequency array for new step");
  f_subject = (double *)mem_alloc(table->number_of_haplo*sizeof(double),"Frequency array for each subject");
  f_0       = (double *)mem_alloc(table->number_of_haplo*sizeof(double),"Frequency array for no association");

  /* Copy genotype and phenotype */
  p = 0;
  for (s=0; s<genotype->Number_of_Subjects; s++) {
    if ((genotype->Subjects + s)->packed_offset >= 0) {
      *(genotype_for_P->Subjects + p) = *(genotype->Subjects + s);
      *(phenotype+p) = (genotype_for_P->Subjects + p)->phenotype;
      p++;
    }
  }
  genotype_for_P->Number_of_Subjects = p;
#ifdef DEBUG_PERMUTATION
  printf("# p = %d : effective subject = %d\n",p,genotype->Number_of_Effective_Subjects);
#endif

  /* Permutation loop start */
  count_upper = 0;
  for (p=0; p<option->permutation; p++) {
    for (s=0; s<genotype_for_P->Number_of_Effective_Subjects; s++) {
      *(random + s) = random_number();
    }
    hpsort_reverse_index_of_double_array(genotype_for_P->Number_of_Effective_Subjects, random, order);

    for (s=0; s<genotype_for_P->Number_of_Effective_Subjects; s++) {
      (genotype_for_P->Subjects + s)->phenotype = *(phenotype + *(order+s));
    }

    /* Set up tables for haplotype inferance */
    table_for_P = setup_EM(genotype_for_P, table_for_P, option->missing_loci);

    /* Make combination table */
    table_for_P->combi = NULL;
    table_for_P->haplo = NULL;
    table_for_P->class = NULL;
    table_for_P = combination(table_for_P);
    all_haplotype(table_for_P);
    set_diplotype_class(table_for_P);
#ifdef DEBUG_PERMUTATION
    dump_estimate_set(table_for_P);
#endif

    /* Initialize probability */
    Init_frequency(table, f_new, option->init_type);
    for (h=0; h<table->number_of_haplo; h++) {
      *(f_0 + h) = *(f_new + h);
    }

    /* EM iteration */
    trait_inheritance_model(table_for_P, table_for_P->model_type, 0);
    likelihood  = EM_iteration_with_penetrance(table_for_P, option, &f_old, &f_new, f_subject, 0);

    /* EM iteration for no association */
    no_association_model(table_for_P);
    likelihood0 = EM_iteration_with_penetrance(table_for_P, option, &f_old, &f_0,   f_subject, 0);
    LRC_max     = likelihood - likelihood0;
    likelihood_max  = likelihood;
    likelihood0_max = likelihood0;
#ifdef DEBUG_PERMUTATION
    printf("# LRC = %e, %e likelihood = %e %e\n",LRC_max,LRC_orig,likelihood,likelihood0);
#endif

    /* Maximum LRC search */
    if (option->maximize > 0) {
      criteria          = log10(option->criteria)*2.0*fabs(LRC_max);
      init_save         = option->init_type;
      option->init_type = 2; /* Random distribution */

      /* Loop for LRC search */
      for (l=1; l<=option->maximize; l++){
	/* Initialize probability */
	Init_frequency(table, f_new, option->init_type);
	for (h=0; h<table->number_of_haplo; h++) {
	  *(f_0 + h) = *(f_new + h);
	}

	/* EM iteration */
	trait_inheritance_model(table_for_P, table_for_P->model_type, 0);
	likelihood  = EM_iteration_with_penetrance(table_for_P, option, &f_old, &f_new, f_subject, 0);

	/* EM iteration for no association */
	no_association_model(table_for_P);
	likelihood0 = EM_iteration_with_penetrance(table_for_P, option, &f_old, &f_0,   f_subject, 0);

	/* Check Likelihood */
	if ((fabs(likelihood - likelihood_max) > criteria) && (likelihood - likelihood_max > criteria)) {
	  likelihood_max  = likelihood;

#ifdef DEBUG_PERMUTATION
	  printf("Restart No. %5d : Likelihood for alternative      = %13.6e\n",l,likelihood);
#endif
	}

	if ((fabs(likelihood0 - likelihood0_max) > criteria) && (likelihood0 - likelihood0_max > criteria)) {
	  likelihood0_max = likelihood0;

#ifdef DEBUG_BOOTSTRAP
	  printf("Restart No. %5d : Likelihood for null             = %13.6e\n",l,likelihood0);
#endif
	}
      }
      LRC_max = likelihood_max - likelihood0_max;
#ifdef DEBUG_PERMUTATION
      printf("  Final Likelihood for alternative hypothesis       = %13.6e\n",likelihood_max);
      printf("  Final Likelihood for null hypothesis              = %13.6e\n",likelihood0_max);
      printf("  Final LRC                                         = %13.6e\n",LRC_max);
#endif

      option->init_type = init_save;
    }

    if (LRC_orig < LRC_max) {
      count_upper++;
    }

    free(table_for_P->class);
    free(table_for_P->haplo);
    free(table_for_P->combi);
    free(table_for_P->geno_info);
  }

  p_value = (double)count_upper/(double)option->permutation;
  if (message) {
    printf("                             p_value = %.4e\n",p_value);
  }
#ifdef DEBUG
  printf("# P value by permutation test = %e,  count = %d\n",p_value,count_upper);
#endif

  free(f_0);
  free(f_subject);
  free(f_new);
  free(f_old);
  free(random);
  free(order);
  free(phenotype);
  free(table_for_P);
  free(genotype_for_P->Subjects);
  free(genotype_for_P);

  return p_value;
}

double         permutation_with_QTL(Genotype_Data *genotype,
				    Estimate_Set  *table,
				    Option_Data   *option,
				    double         LRC_orig,
				    int            message)
{
  /* Local variables */
  double       p_value;
  int          s, p, l, count_upper, init_save;
  int         *order;
  double      *phenotype, *random, *f_new, *f_old, *f_subject, *f_0;
  double       likelihood, likelihood0,  LRC_max, criteria;
  double       likelihood_max, likelihood0_max;
  haplotype    h;

  /* Data structure for permutation */
  Genotype_Data *genotype_for_P;
  Estimate_Set  *table_for_P;

  /* Message */
  if (message) {
    printf("\nPermutation test : number of samples = %d\n",
	   option->permutation);
  }

  /* Allocate data structure */
  genotype_for_P           = (Genotype_Data *)mem_alloc(sizeof(Genotype_Data),"Genotype data structure");
  *genotype_for_P          = *genotype;
  genotype_for_P->Subjects = (Subject_Genotype *)mem_alloc(genotype->Number_of_Effective_Subjects*sizeof(Subject_Genotype),"Subject array");

  table_for_P              = alloc_estimate_set();
  *table_for_P             = *table;

#ifdef DEBUG_PERMUTATION
  dump_estimate_set(table_for_P);
#endif

  /* Allocate data area */
  phenotype = (double *)mem_alloc(genotype->Number_of_Effective_Subjects*sizeof(double),"Phenotype array for permutation");
  order     = (int *)   mem_alloc(genotype->Number_of_Effective_Subjects*sizeof(int),   "Index array for phenotype");
  random    = (double *)mem_alloc(genotype->Number_of_Effective_Subjects*sizeof(double),"Array to randomize phenotype");
  f_old     = (double *)mem_alloc(table->number_of_haplo*sizeof(double),"Frequency array for old step");
  f_new     = (double *)mem_alloc(table->number_of_haplo*sizeof(double),"Frequency array for new step");
  f_subject = (double *)mem_alloc(table->number_of_haplo*sizeof(double),"Frequency array for each subject");
  f_0       = (double *)mem_alloc(table->number_of_haplo*sizeof(double),"Frequency array for no association");

  /* Copy genotype and phenotype */
  p = 0;
  for (s=0; s<genotype->Number_of_Subjects; s++) {
    if ((genotype->Subjects + s)->packed_offset >= 0) {
      *(genotype_for_P->Subjects + p) = *(genotype->Subjects + s);
      *(phenotype+p) = (genotype_for_P->Subjects + p)->QTL;
      p++;
    }
  }
#ifdef DEBUG_PERMUTATION
  printf("# p = %d : effective subject = %d\n",p,genotype->Number_of_Effective_Subjects);
#endif

  /* Permutation loop start */
  count_upper = 0;
  for (p=0; p<option->permutation; p++) {
    for (s=0; s<genotype_for_P->Number_of_Effective_Subjects; s++) {
      *(random + s) = random_number();
    }
    hpsort_reverse_index_of_double_array(genotype_for_P->Number_of_Effective_Subjects, random, order);

    for (s=0; s<genotype_for_P->Number_of_Effective_Subjects; s++) {
      (genotype_for_P->Subjects + s)->QTL = *(phenotype + *(order+s));
    }

    /* Set up tables for haplotype inferance */
    table_for_P = setup_EM(genotype_for_P, table_for_P, option->missing_loci);

    /* Make combination table */
    table_for_P->combi = NULL;
    table_for_P->haplo = NULL;
    table_for_P->class = NULL;
    table_for_P = combination(table_for_P);
    all_haplotype(table_for_P);
    set_diplotype_class(table_for_P);
#ifdef DEBUG_PERMUTATION
    dump_estimate_set(table_for_P);
#endif

    /* Initialize probability */
    Init_frequency(table, f_new, option->init_type);
    for (h=0; h<table->number_of_haplo; h++) {
      *(f_0 + h) = *(f_new + h);
    }

    /* EM iteration */
    trait_inheritance_model(table_for_P, table_for_P->model_type, 0);
    likelihood  = EM_iteration_with_QTL(table_for_P, option, genotype_for_P, &f_old, &f_new, f_subject, 0);

    /* EM iteration for no association */
    no_association_model(table_for_P);
    likelihood0 = EM_iteration_with_QTL(table_for_P, option, genotype_for_P, &f_old, &f_0,   f_subject, 0);
    LRC_max     = likelihood - likelihood0;
    likelihood_max  = likelihood;
    likelihood0_max = likelihood0;
#ifdef DEBUG_PERMUTATION
    printf("# LRC = %e, %e likelihood = %e %e\n",LRC_max,LRC_orig,likelihood,likelihood0);
#endif

    /* Maximum LRC search */
    if (option->maximize > 0) {
      criteria          = log10(option->criteria)*2.0*fabs(LRC_max);
      init_save         = option->init_type;
      option->init_type = 2; /* Random distribution */

      /* Loop for LRC search */
      for (l=1; l<=option->maximize; l++){
	/* Initialize probability */
	Init_frequency(table, f_new, option->init_type);
	for (h=0; h<table->number_of_haplo; h++) {
	  *(f_0 + h) = *(f_new + h);
	}

	/* EM iteration */
	trait_inheritance_model(table_for_P, table_for_P->model_type, 0);
	likelihood  = EM_iteration_with_QTL(table_for_P, option, genotype_for_P, &f_old, &f_new, f_subject, 0);

	/* EM iteration for no association */
	no_association_model(table_for_P);
	likelihood0 = EM_iteration_with_QTL(table_for_P, option, genotype_for_P, &f_old, &f_0,   f_subject, 0);

	/* Check Likelihood */
	if ((fabs(likelihood - likelihood_max) > criteria) && (likelihood - likelihood_max > criteria)) {
	  likelihood_max  = likelihood;

#ifdef DEBUG_PERMUTATION
	  printf("Restart No. %5d : Likelihood for alternative      = %13.6e\n",l,likelihood);
#endif
	}

	if ((fabs(likelihood0 - likelihood0_max) > criteria) && (likelihood0 - likelihood0_max > criteria)) {
	  likelihood0_max = likelihood0;

#ifdef DEBUG_BOOTSTRAP
	  printf("Restart No. %5d : Likelihood for null             = %13.6e\n",l,likelihood0);
#endif
	}
      }
      LRC_max = likelihood_max - likelihood0_max;
#ifdef DEBUG_PERMUTATION
      printf("  Final Likelihood for alternative hypothesis       = %13.6e\n",likelihood_max);
      printf("  Final Likelihood for null hypothesis              = %13.6e\n",likelihood0_max);
      printf("  Final LRC                                         = %13.6e\n",LRC_max);
#endif

      option->init_type = init_save;
    }

    if (LRC_orig < LRC_max) {
      count_upper++;
    }

    free(table_for_P->class);
    free(table_for_P->haplo);
    free(table_for_P->combi);
    free(table_for_P->geno_info);
  }

  p_value = (double)count_upper/(double)option->permutation;
  if (message) {
    printf("                             p_value = %.4e\n",p_value);
  }
#ifdef DEBUG
  printf("# P value by permutation test = %e,  count = %d\n",p_value,count_upper);
#endif

  free(f_0);
  free(f_subject);
  free(f_new);
  free(f_old);
  free(random);
  free(order);
  free(phenotype);
  free(table_for_P);
  free(genotype_for_P->Subjects);
  free(genotype_for_P);

  return p_value;
}
