#include "haplotypetest.h"

/* Utility function */
extern void          *mem_alloc(size_t, const char *);
extern void          *mem_realloc(void *, size_t, const char *);

/* Error handler */
extern int            error_check();
extern int            error_message(char *);
extern int            warning_message(char *);

/* Define functions */
Estimate_Set  *select_loci(Genotype_Data *, Option_Data *);
Estimate_Set  *select_window(int, int *, Genotype_Data *, Option_Data *);
int           *filter_loci(Genotype_Data *, Option_Data *, int *);
Estimate_Set  *alloc_estimate_set();
void           free_estimate_set(Estimate_Set *);
void           save_non_zero_haplotype_set(Estimate_Set *, Option_Data *);

Estimate_Set  *select_loci(Genotype_Data *genotype, Option_Data *option)
{
  /* Local variables */
  Estimate_Set   *table_for_EM;
  int             l, i, count;
  double          criteria, df;

  /* Memory allocation for the problem definition table */
  table_for_EM = alloc_estimate_set();

#ifdef DEBUG
  printf("\n# SNPs frequency for rejection = %6.3f\n",option->SNP_rejection);
#endif

  /* 1st : Select loci according to input data and allele frequency */
  criteria = 1.0-option->SNP_rejection;
  for (l=0; l<genotype->Number_of_Loci; l++) {
    (genotype->Locus_List + l)->filtered = FILTERED_LOCUS;
  }

  count = 0;
  if (option->select_loci > 0) {
    for (i=0; i<option->select_loci; i++) {
      l = *(option->loci + i)-1;
      if (l >= genotype->Number_of_Loci) {
	sprintf(error_message_buffer,"Locus %d is selected by option, but is not in genotype file.",l);
	warning_message(error_message_buffer);
      }
      else {
	if ((genotype->Locus_List+l)->Allele_Frequency <= criteria) {
	  (genotype->Locus_List+l)->filtered = count;
	  count++;
	}
      }
    }
  }
  else {
    for (l=0; l<genotype->Number_of_Loci; l++) {
      if ((genotype->Locus_List+l)->Allele_Frequency <= criteria) {
	(genotype->Locus_List+l)->filtered = count;
	count++;
      }
    }
  }

  /* 2nd : Check errors */
  if      (count == 0) {
    error_message("No loci fit the SNP rejection criteria and input selection.");
    error_check();
  }
  else if (count > MAX_NUMBER_OF_LOCI) {
    sprintf(error_message_buffer,"The number of selected loci exceeds program capability of %d.",MAX_NUMBER_OF_LOCI);
    error_message(error_message_buffer);
  }
  error_check();

  /* 3rd : Set the selected loci number to the table */
  table_for_EM->select_loci = count;
  table_for_EM->loci        = (int *)   mem_alloc(count*sizeof(int),   "Loci selection table");
  table_for_EM->f_allele    = (double *)mem_alloc(count*sizeof(double),"Allele frequency for selected loci");

#ifdef DEBUG
  printf("# Number of seletced loci = %d\n",table_for_EM->select_loci);
  for (l=0; l<genotype->Number_of_Loci; l++) {
    printf("# Locus %5d : flag = %5d",l+1,(genotype->Locus_List + l)->filtered);
    if ((genotype->Locus_List + l)->filtered != FILTERED_LOCUS) {
      printf(" selected");
    }
    printf("\n");
  }
#endif

  count = 0;
  df    = 1.0;
  table_for_EM->df = 0;
  for (l=0; l<genotype->Number_of_Loci; l++) {
    if ((genotype->Locus_List + l)->filtered != FILTERED_LOCUS) {
      *(table_for_EM->loci     + count) = l;
      *(table_for_EM->f_allele + count) = (genotype->Locus_List + l)->Allele_Frequency;
      df = df * (double)(genotype->Locus_List + l)->number_of_allele;
      table_for_EM->df = table_for_EM->df + (genotype->Locus_List + l)->number_of_allele;
      count++;
    }
  }
  df    = df - 1.0;
  table_for_EM->df = (int)df - table_for_EM->df + table_for_EM->select_loci;

  printf("\nNumber of loci to infer haplotype is %d\n",table_for_EM->select_loci);
  printf("  Loci number :");
  for (l=0; l<table_for_EM->select_loci; l++) {
    printf(" %d",*(table_for_EM->loci + l)+1);
  }
  printf("\n");

  table_for_EM->target           = 0;
  table_for_EM->target_mask      = 0;
  table_for_EM->df_for_test      = 0;

  return table_for_EM;
}

Estimate_Set  *select_window(int            start,
			     int           *loci_list,
			     Genotype_Data *genotype, 
			     Option_Data   *option)
{
  /* Local variables */
  Estimate_Set   *table_for_EM;
  int             l, i;
  double          df;

  /* Memory allocation for the problem definition table */
  table_for_EM = alloc_estimate_set();

  /* Set the selected loci number to the table */
  table_for_EM->select_loci = option->window_size;
  table_for_EM->loci        = (int *)   mem_alloc(option->window_size*sizeof(int),   "Loci selection table");
  table_for_EM->f_allele    = (double *)mem_alloc(option->window_size*sizeof(double),"Allele frequency for selected loci");

  df    = 1.0;
  table_for_EM->df = 0;
  for (i=0; i<table_for_EM->select_loci; i++) {
    l = i + start;
    *(table_for_EM->loci     + i) = *(loci_list + l);
    *(table_for_EM->f_allele + i) = (genotype->Locus_List + *(loci_list + l))->Allele_Frequency;
    df = df * (double)(genotype->Locus_List + *(loci_list + l))->number_of_allele;
    table_for_EM->df = table_for_EM->df + (genotype->Locus_List + *(loci_list + l))->number_of_allele;
  }
#ifdef DEBUG
  printf("# Number of seletced loci = %d\n",table_for_EM->select_loci);
  for (i=0; i<table_for_EM->select_loci; i++) {
    printf("# Locus %5d : flag = %5d selected.\n",*(table_for_EM->loci+i)+1,(genotype->Locus_List + *(table_for_EM->loci+i))->filtered);
  }
#endif

  df    = df - 1.0;
  table_for_EM->df = (int)df - table_for_EM->df + table_for_EM->select_loci;
  table_for_EM->target           = 0;
  table_for_EM->target_mask      = 0;
  table_for_EM->df_for_test      = 0;

  return table_for_EM;
}

int           *filter_loci(Genotype_Data *genotype,
			   Option_Data   *option,
			   int           *count)
{
  /* Local variables */
  int             l, *loci_list, *lptr;
  double          criteria;

  /* 1st : Select loci according to allele frequency */
  criteria = 1.0-option->SNP_rejection;
  *count   = 0;
  for (l=0; l<genotype->Number_of_Loci; l++) {
    if ((genotype->Locus_List+l)->Allele_Frequency < criteria) {
      (genotype->Locus_List+l)->filtered = *count;
      *count += 1;
    }
    else {
      (genotype->Locus_List+l)->filtered = FILTERED_LOCUS;
    }
  }

  loci_list = (int *)mem_alloc(*count*sizeof(int),"Non-filtered loci table");
  lptr      = loci_list;
  for (l=0; l<genotype->Number_of_Loci; l++) {
    if ((genotype->Locus_List+l)->filtered != FILTERED_LOCUS) {
      *lptr = l; lptr++;
    }
  }

  return loci_list;
}

Estimate_Set  *alloc_estimate_set()
{
  /* Local variables */
  Estimate_Set   *table_for_EM;

  /* Memory allocation for the problem definition table */
  table_for_EM = (Estimate_Set *)mem_alloc(sizeof(Estimate_Set),"Problem definition table");
  table_for_EM->loci      = NULL;
  table_for_EM->f_allele  = NULL;
  table_for_EM->geno_info = NULL;
  table_for_EM->combi     = NULL;
  table_for_EM->class     = NULL;
  table_for_EM->haplo     = NULL;
  table_for_EM->order     = NULL;
  table_for_EM->f         = NULL;

  return table_for_EM;
}

void           free_estimate_set(Estimate_Set *table)
{
  if (table->f != NULL) {
    free(table->f);
  }
  if (table->order != NULL) {
    free(table->order);
  }
  if (table->haplo != NULL) {
    free(table->haplo);
  }
  if (table->class != NULL) {
    free(table->class);
  }
  if (table->combi != NULL) {
    free(table->combi);
  }
  if (table->geno_info != NULL) {
    free(table->geno_info);
  }
  if (table->f_allele != NULL) {
    free(table->f_allele);
  }
  if (table->loci != NULL) {
    free(table->loci);
  }
  free(table);

  return;
}

void           save_non_zero_haplotype_set(Estimate_Set *table, Option_Data *option)
{
  int         h;
  double     *f;
  haplotype  *haplo, *order, nonzero;

  if (table->class != NULL) {
    free(table->class); table->class = NULL;
  }
  if (table->combi != NULL) {
    free(table->combi); table->combi = NULL;
  }
  if (table->geno_info != NULL) {
    free(table->geno_info); table->geno_info = NULL;
  }
  if (table->f_allele != NULL) {
    free(table->f_allele); table->f_allele = NULL;
  }

  nonzero = 0;
  if (option->cut_frequency == CUT_FREQUENCY) {
    option->cut_frequency = table->normal * 5.0;
  }
  for (h=0; h<table->nonzero; h++) {
    if (*(table->f + *(table->order + h)) < option->cut_frequency) {
      break;
    }
    nonzero++;
#ifdef DEBUG
    printf("# %2d %7.4f %x\n",h+1,*(table->f + *(table->order + h)),*(table->haplo + *(table->order + h)));
#endif
  }

  if (nonzero > MAX_NUMBER_OF_LOCI) {
    sprintf(error_message_buffer,"Number of non zero haplotype %d exceeds the limit. Reset to %d.",nonzero,MAX_NUMBER_OF_LOCI);
    warning_message(error_message_buffer);
    nonzero = MAX_NUMBER_OF_LOCI;
  }
  table->nonzero = nonzero;
  f     = (double *)mem_alloc(table->nonzero*sizeof(double),"Frequency array for haplotype");
  haplo = (haplotype *)mem_alloc(table->nonzero*sizeof(double),"Non zero haplotype array");
  order = (haplotype *)mem_alloc(table->nonzero*sizeof(double),"Index array for non zero haplotype array");

  for (h=0; h<table->nonzero; h++) {
    *(f    +h) = *(table->f     + *(table->order + h));
    *(haplo+h) = *(table->haplo + *(table->order + h));
    *(order+h) = h;
#ifdef DEBUG
    printf("# %2d %7.4f %x\n",h+1,*(f+h),*(haplo+h));
#endif
  }

  if (table->f != NULL) {
    free(table->f);     table->f = f;
  }
  if (table->order != NULL) {
    free(table->order); table->order = order;
  }
  if (table->haplo != NULL) {
    free(table->haplo); table->haplo = haplo;
  }

  return;
}
